import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
import matplotlib.colors
from jinja2 import Environment, FileSystemLoader

from ficture.utils.visualize_factors import visual_hc, image_to_base64

def factor_report(_args):

    parser = argparse.ArgumentParser(prog="factor_report")
    parser.add_argument('--path', type=str, help='')
    parser.add_argument('--pref', type=str, help='')
    parser.add_argument('--model_id', type=str, default='', help='')
    parser.add_argument('--color_table', type=str, default='', help='')
    parser.add_argument('--n_top_gene', type=int, default=20, help='')
    parser.add_argument('--min_top_gene', type=int, default=10, help='')
    parser.add_argument('--max_pval', type=float, default=0.001, help='')
    parser.add_argument('--min_fc', type=float, default=1.5, help='')
    parser.add_argument('--output_pref', type=str, default='', help='')

    parser.add_argument('--hc_tree', action='store_true')
    parser.add_argument('--n_top_gene_on_tree', type=int, default=10, help='')
    parser.add_argument('--tree_figure', type=str, default='', help='')
    parser.add_argument('--cprob_cut', type=str, default='.99', help='Only visualize top factors with cumulative probability > cprob_cut')
    parser.add_argument('--model', type=str, default='', help='')
    parser.add_argument('--circle_if', type=int, default=24, help='')
    parser.add_argument('--remake_tree', action='store_true')
    parser.add_argument('--circle', action='store_true')
    parser.add_argument('--vertical', action='store_true')
    args = parser.parse_args(_args)

    path=args.path
    pref=args.pref
    ntop = args.n_top_gene
    mtop = args.min_top_gene
    pval_max = args.max_pval
    fc_min = args.min_fc
    ejs = os.path.join(os.path.dirname(__file__), "factor_report.template.html")
    if not os.path.isfile(ejs):
        sys.exit(f"Template file {ejs} not found")
    if args.remake_tree or args.circle or args.vertical:
        args.hc_tree = True

    model_id = args.model_id
    if model_id == '':
        model_id = args.pref

    output_pref = args.output_pref
    if output_pref == '':
        output_pref = path+"/"+pref

    # Color code
    color_f = args.color_table
    if os.path.isfile(args.color_table):
        color_table = pd.read_csv(args.color_table, sep='\t')
    else:
        color_f = path+"/figure/"+model_id+".rgb.tsv"
        color_table = pd.read_csv(color_f, sep='\t')
    K = color_table.shape[0]
    factor_header = np.arange(K).astype(str)
    color_table.Name = color_table.Name.astype(int)
    color_table.sort_values(by = 'Name', inplace=True)
    color_table.index = color_table.Name.values
    color_table['RGB'] = [','.join(x) for x in np.clip((color_table.loc[:, ['R','G','B']].values * 255).astype(int), 0, 255).astype(str) ]
    color_table['HEX'] = [ matplotlib.colors.to_hex(v) for v in np.array(color_table.loc[:, ['R','G','B']]) ]
    node_color = {str(i):v['HEX'] for i,v in color_table.iterrows() }


    # Posterior count
    f=path+"/"+pref+".posterior.count.tsv.gz"
    post = pd.read_csv(f, sep='\t')
    recol = {}
    for u in post.columns:
        v = re.match('^[A-Za-z]*_*(\d+)$', u.strip())
        if v:
            recol[v.group(0)] = v.group(1)
    post.rename(columns=recol, inplace=True)
    post_umi = post.loc[:, factor_header].sum(axis = 0).astype(int).values
    post_weight = post.loc[:, factor_header].sum(axis = 0).values
    post_weight /= post_weight.sum()

    # DE genes
    f=path+"/DE/"+pref+".bulk_chisq.tsv"
    de = pd.read_csv(f, sep='\t')
    de.factor = de.factor.astype(int)
    top_gene = []
    top_gene_anno = []
    de['Rank'] = 0
    # Temporary: shorten unspliced gene names
    de.gene = de.gene.str.replace('unspl_', 'u_')
    # Top genes by Chi2
    de.sort_values(by=['factor','Chi2'],ascending=False,inplace=True)
    for k in range(K):
        v = de.loc[de.factor.eq(k), 'gene'].iloc[:args.n_top_gene_on_tree].values
        top_gene_anno.append(', '.join(v))
        de.loc[de.factor.eq(k), 'Rank'] = np.arange(de.factor.eq(k).sum())
        v = de.loc[de.factor.eq(k) & ( (de.Rank < mtop) | \
                ((de.pval <= pval_max) & (de.FoldChange >= fc_min)) ), \
                'gene'].iloc[:ntop].values
        if len(v) == 0:
            top_gene.append([k, '.'])
        else:
            top_gene.append([k, ', '.join(v)])
    # Top genes by fold change
    de.sort_values(by=['factor','FoldChange'],ascending=False,inplace=True)
    for k in range(K):
        de.loc[de.factor.eq(k), 'Rank'] = np.arange(de.factor.eq(k).sum())
        v = de.loc[de.factor.eq(k) & ( (de.Rank < mtop) | \
                ((de.pval <= pval_max) & (de.FoldChange >= fc_min)) ), \
                'gene'].iloc[:ntop].values
        if len(v) == 0:
            top_gene[k].append('.')
        else:
            top_gene[k].append(', '.join(v))
    # Top genes by absolute weight
    for k in range(K):
        v = post.gene.iloc[np.argsort(post.loc[:, str(k)].values)[::-1][:ntop] ].values
        top_gene[k].append(', '.join(v))

    # Summary
    table = pd.DataFrame({'Factor':np.arange(K), 'RGB':color_table.RGB.values,
                        'Weight':post_weight, 'PostUMI':post_umi,
                        'TopGene_pval':[x[1] for x in top_gene],
                        'TopGene_fc':[x[2] for x in top_gene],
                        'TopGene_weight':[x[3] for x in top_gene] })
    table.sort_values(by = 'Weight', ascending = False, inplace=True)

    f = output_pref+".factor.info.tsv"
    table.to_csv(f, sep='\t', index=False, header=True, float_format="%.5f")

    f = output_pref+".factor.info.tsv"
    with open(f, 'r') as rf:
        lines = rf.readlines()
    header = lines[0].strip().split('\t')
    rows = [ list(enumerate(row.strip().split('\t') )) for row in lines[1:]]

    # Load template
    env = Environment(loader=FileSystemLoader(os.path.dirname(ejs)))
    template = env.get_template(os.path.basename(ejs))

    image_base64 = None
    tree_alt = None
    tree_caption = None

    if args.hc_tree:
        # Hierarchical clustering
        m = re.match("^[0\.]*(\d+)$", args.cprob_cut)
        if m is None:
            sys.exit(f"Invalid --cprob_cut, please use a number between 0 and 1 (e.g. 0.99)")
        cprob_label = m.group(1)
        cprob_cut  = float(args.cprob_cut)

        tree_f = args.tree_figure
        if not os.path.exists(args.tree_figure) or args.remake_tree:
            tree_f = os.path.dirname(color_f) + '/' + pref + ".coshc."+cprob_label+".tree.png"
            if not os.path.exists(tree_f) or args.remake_tree:
                model_f = args.model
                if not os.path.exists(args.model):
                    model_f = path + "/" + model_id + ".model_matrix.tsv.gz"
                if not os.path.exists(model_f):
                    print("Cannot find model file, will cluster based on posterior count")
                    model = post
                else:
                    model = pd.read_csv(model_f, sep='\t')
                model_prob = np.array(model.iloc[:, 1:]).T + .1
                model_prob = model_prob / model_prob.sum(axis = 1).reshape((-1,1))

                circle = args.circle
                if not circle and args.circle_if > 0:
                    v = np.argsort(post_weight)[::-1]
                    w = np.cumsum(post_weight[v] )
                    k = np.arange(K)[w > cprob_cut][0]
                    if k > args.circle_if:
                        circle = True
                tree = visual_hc(model_prob, post_weight, top_gene_anno, \
                                node_color = node_color, circle = circle, \
                                output_f = tree_f, cprob_cut = cprob_cut)
        print(f"Tree figure path: {tree_f}")
        image_base64 = image_to_base64(tree_f)

        tree_alt = "Hierarchical clustering of factors based on pairwise coine distance"
        tree_caption = "Clustering of factors based on pairwise coine distance. Factors with high abundance jointly accounting for " + args.cprob_cut + " of observations are displayed."

    # Render the HTML file
    html_output = template.render(header=header, rows=rows, image_base64=image_base64, tree_image_alt=tree_alt, tree_image_caption=tree_caption)

    f=output_pref+".factor.info.html"
    with open(f, "w") as html_file:
        html_file.write(html_output)

    print(f)



# Visualize pixel level factor analysis results
# <= 5 (preferably 3) factors at a time, with specified colors
# Input file contains only top k factors and probabilities per pixel
# Meant to make use of the indexed input to plot for specified regions quickly

import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
from scipy.sparse import *
import subprocess as sp
from joblib.parallel import Parallel, delayed
import matplotlib as mpl
import cv2

from ficture.utils import utilt
from ficture.loaders.pixel_factor_loader import BlockIndexedLoader

def plot_pixel_multi(_args):

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='Output file full path')
    parser.add_argument('--channel_list', type=str, nargs='+', help="Select up to 3 channels/factors to visualize")
    parser.add_argument('--color_list', type=str, nargs='*', default=["#144A74", "#FF9900", "#FFEC11", "#DD65E6"], help='')
    parser.add_argument('--color_table', type=str, default='', help='Pre-defined color map')
    parser.add_argument('--color_table_index_column', type=str, default='Name', help='')
    parser.add_argument('--xmin', type=float, default=-np.inf, help="")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="")
    parser.add_argument('--xmax', type=float, default=np.inf, help="")
    parser.add_argument('--ymax', type=float, default=np.inf, help="")
    parser.add_argument('--full', action='store_true', help="")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Actual size (um) corresponding to each pixel in the output image")
    parser.add_argument('--categorical', action='store_true', help="Plot top factor for each pixel categorically, without probability or mixture")
    parser.add_argument('--debug', action='store_true', help="")

    args = parser.parse_args(_args)
    logger =  logging.getLogger(__name__)

    rgb=list("RGB")
    bgr=list("BGR") # opencv default channel order
    bgra=list("BGRA") # opencv bgra
    pcut = 0.01
    spcut = 0.1
    if len(args.channel_list) > 5:
        logger.error("Be colorblind friendly, visualize <=5 and preferably 3 channels at a time")
        sys.exit(1)
    channels = args.channel_list
    # Read color table
    if os.path.exists(args.color_table):
        color_info = pd.read_csv(args.color_table, sep='\t', header=0, index_col=args.color_table_index_column)
        if color_info[rgb].max().max() > 1:
            logger.warning("Color table should contain RGB values in 0-1 range, but found values > 1, will interpret as in 0-255 range")
            for c in rgb:
                color_info[c] = np.clip(color_info[c]/255,0,1)
        color_info.index = color_info.index.astype(str)
        channels = [x for x in channels if x in color_info.index]
        if len(channels) != len(args.channel_list):
            logger.error("Some channels are not found in the input color table")
            sys.exit(1)
        color_info = {i: np.array(color_info.loc[i, rgb]) for i in channels}
    else:
        if len(args.color_list) < len(channels):
            logger.error("Number of input colors and channels do not match")
            sys.exit(1)
        color_list = [[int(x.strip('#')[i:i+2],16)/255 for i in (0,2,4)] for x in args.color_list]
        color_info = {x:np.array(color_list[i]) for i,x in enumerate(channels)}

    logger.info(f"Read color map {color_info}")

    loader = BlockIndexedLoader(args.input, args.xmin, args.xmax, args.ymin, args.ymax, args.full)
    width = int((loader.xmax - loader.xmin + 1)/args.plot_um_per_pixel)
    height= int((loader.ymax - loader.ymin + 1)/args.plot_um_per_pixel)
    logger.info(f"Image size {height} x {width}")

    # Read input file, fill the rgb matrix
    img = np.zeros((height,width,4), dtype=np.uint8)
    img[:,:,3] = 255
    for df in loader:
        if df.shape[0] == 0:
            continue
        logger.info(f"Reading pixels... {df.X.iloc[-1]}, {df.Y.iloc[-1]}, {df.shape[0]}")
        indx = np.zeros(df.shape[0], dtype=bool)
        if args.categorical:
            indx = df.K1.isin(channels)
        else:
            for k in range(1, loader.meta['TOPK']):
                indx = indx | (df[f"K{k}"].isin(channels) & df[f"P{k}"].gt(pcut))
        df = df.loc[indx, :]
        if df.shape[0] == 0:
            continue
        df['X'] = np.clip(((df.X - loader.xmin) / args.plot_um_per_pixel).astype(int),0,width-1)
        df['Y'] = np.clip(((df.Y - loader.ymin) / args.plot_um_per_pixel).astype(int),0,height-1)
        df[bgra] = 0
        if args.categorical:
            df.loc[:,rgb] = np.array(list(df.K1.map(color_info)) )
            df['A'] = 1
        else:
            for k in range(1, loader.meta['TOPK']):
                indx = df[f"K{k}"].isin(channels)
                if indx.sum() == 0:
                    continue
                df.loc[indx, rgb] = np.array(list(df.loc[indx, f"K{k}"].map(color_info))) * df.loc[indx, f"P{k}"].values.reshape((-1,1))
                df.loc[indx, 'A'] += df.loc[indx, f"P{k}"]
        df = df.groupby(by=['X','Y']).agg({c:np.mean for c in bgra}).reset_index()
        df = df.loc[df.A.gt(spcut), :]
        df[bgra] = np.clip(np.around(df[bgra] * 255),0,255).astype(np.uint8)
        for i,c in enumerate(bgra):
            img[df.Y.values, df.X.values, [i]*df.shape[0]] = df[c].values
        if args.debug:
            break

    if not args.output.endswith(".png"):
        args.output += ".png"
    cv2.imwrite(args.output,img)
    logger.info(f"Finished")



'''
Transform into factor space based on input model
Model contains gene names and either Dirichlet parameters (or probabilities for more general use?)
Input pixel level data will be grouped into (overlapping) hexagons
'''
import sys, os, copy, gzip, time, logging, pickle, argparse
import numpy as np
import pandas as pd
import random as rng

from ficture.loaders.pixel_to_unit_loader import PixelToUnit
from ficture.models.online_lda import OnlineLDA

def transform(_args):

    parser = argparse.ArgumentParser(prog = "transform")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output_pref', type=str, help='')
    parser.add_argument('--model', type=str, help='')

    parser.add_argument('--key', type=str, default = 'gn', help='gt: genetotal, gn: gene, spl: velo-spliced, unspl: velo-unspliced')
    parser.add_argument('--major_axis', type=str, default=None, help='X or Y')
    parser.add_argument('--region_id', type=str, default=None, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=20, help='')
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--thread', type=int, default=-1, help='')
    parser.add_argument('--n_move', type=int, default=3, help='')
    parser.add_argument('--hex_width', type=int, default=24, help='')
    parser.add_argument('--hex_radius', type=int, default=-1, help='')
    parser.add_argument('--precision', type=int, default=1, help='Number of digits to store spatial location (in um), 0 for integer.')
    parser.add_argument('--chunksize', type=int, default=100000, help='Number of lines to read at a time')

    args = parser.parse_args(_args)
    logger = logging.getLogger(__name__)
    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file.")
    if not os.path.exists(args.model):
        sys.exit("ERROR: cannot find model file.")
    if args.hex_width <= 0 and args.hex_radius <= 0:
        sys.exit("ERROR: invalid hex_width or hex_radius")
    if not os.path.exists(os.path.dirname(args.output_pref)):
        os.makedirs(os.path.dirname(args.output_pref))

    ### Input
    with gzip.open(args.input, 'rt') as rf:
        header=rf.readline().strip().split('\t')
    key = args.key.lower()
    header=[x.upper() if len(x) == 1 else x.lower() for x in header]
    usecol = ['X','Y','gene',key]
    adt={'X':float,'Y':float,'gene':str,args.key:int}
    if args.region_id is not None:
        usecol.append(args.region_id)
        adt[args.region_id] = str
    miss_header = [x for x in usecol if x not in header]
    if len(miss_header) > 0:
        sys.exit(f"ERROR: {miss_header} is not in the input file")
    reader = pd.read_csv(gzip.open(args.input, 'rt'), sep='\t',\
                        chunksize=args.chunksize, skiprows=1, names=header,\
                        usecols=usecol, dtype=adt)

    ### Model
    if args.model.endswith('.tsv.gz') or args.model.endswith('tsv'):
        model_mtx = pd.read_csv(args.model, sep='\t')
        feature_kept=list(model_mtx.gene)
        model_mtx = np.array(model_mtx.iloc[:, 1:])
        M, K = model_mtx.shape
        model = OnlineLDA(vocab=feature_kept,K=K,N=1e4,thread=args.thread,tol=1e-3)
        model.init_global_parameter(model_mtx.T)
    else:
        try:
            model = pickle.load(open(args.model, 'rb'))
            model.n_jobs = args.thread
            feature_kept = model.feature_names_in_
            K, M = model.components_.shape
            model.feature_names_in_ = None
        except:
            sys.exit("ERROR: --model must be either a tsv file containing gene name and factor-gene loadings or a pickle model object from sklearn LDA")
    ft_dict = {x:i for i,x in enumerate(feature_kept)}
    logger.info(f"Model loaded with {M} features and {K} factors")

    ### Basic parameterse
    factor_header = [str(x) for x in range(K)]
    radius = args.hex_radius
    if radius < 0:
        radius = args.hex_width / np.sqrt(3)
    b_size = radius * 10

    # Pixel reader
    batch_obj = PixelToUnit(reader, ft_dict, key, radius,\
                scale=1./args.mu_scale, min_ct_per_unit=args.min_ct_per_unit,\
                sliding_step=args.n_move, major_axis=args.major_axis)

    # Transform
    post_count = np.zeros((K, M))
    n_unit = 0
    n_batch= 0
    oheader = ["unit",key,"x","y","topK","topP"]+factor_header
    out_f = args.output_pref + ".fit_result.tsv.gz"
    with gzip.open(out_f, 'wt') as wf:
        wf.write('\t'.join(oheader) + '\n')
    t0 = time.time()
    while batch_obj.read_chunk(min_size=b_size):
        theta = model.transform(batch_obj.mtx)
        post_count += np.array(theta.T @ batch_obj.mtx)
        n_batch += 1
        n_unit  += theta.shape[0]
        t1 = time.time() - t0
        batch_obj.brc['topK'] = np.argmax(theta, axis = 1)
        batch_obj.brc['topP'] = theta.max(axis = 1)
        batch_obj.brc = pd.concat([batch_obj.brc, pd.DataFrame(theta, columns=factor_header)], axis=1)
        batch_obj.brc['x'] = batch_obj.brc.x.map(lambda x : f"{x:.{args.precision}f}")
        batch_obj.brc['y'] = batch_obj.brc.y.map(lambda x : f"{x:.{args.precision}f}")
        batch_obj.brc.rename(columns = {'hex_id':'unit'}, inplace=True)
        batch_obj.brc[oheader].to_csv(out_f, sep='\t', index=False, header=False, float_format='%.3e', mode='a', compression={"method":"gzip"})
        logger.info(f"Transformed {n_batch} batches with total {n_unit} units, {t1/60:2f}min")

    out_f = args.output_pref + ".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({'gene': feature_kept}),\
            pd.DataFrame(post_count.T, dtype='float64',\
                            columns = [str(k) for k in range(K)])],\
                axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s:%(levelname)s: %(message)s", "%I:%M:%S %p")
handler.setFormatter(formatter)
logger.addHandler(handler)



# Visualize pixel level factor analysis results
# Input file contains only top k factors and probabilities per pixel
# Meant to make use of the indexed input to plot for specified regions quickly

import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
from scipy.sparse import *
import subprocess as sp
from joblib.parallel import Parallel, delayed
import matplotlib as mpl
import cv2

from ficture.loaders.pixel_factor_loader import BlockIndexedLoader

def plot_pixel_full(_args):

    parser = argparse.ArgumentParser(prog="plot_pixel_full")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='Output file full path')
    parser.add_argument('--color_table', type=str, default='', help='Pre-defined color map')
    parser.add_argument('--color_table_index_column', type=str, default='Name', help='')
    parser.add_argument('--xmin', type=float, default=-np.inf, help="")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="")
    parser.add_argument('--xmax', type=float, default=np.inf, help="")
    parser.add_argument('--ymax', type=float, default=np.inf, help="")
    parser.add_argument('--full', action='store_true', help="Read full input")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Actual size (um) corresponding to each pixel in the output image")
    parser.add_argument('--plot_top', action='store_true', help="Plot top factor only")

    args = parser.parse_args(_args)
    logger =  logging.getLogger(__name__)

        # Read color table
    rgb=['B','G','R'] # opencv rgb order
    cdty = {x:float for x in rgb}
    color_info = pd.read_csv(args.color_table, sep='\t', header=0, index_col=args.color_table_index_column, dtype=cdty)
    color_info.index = color_info.index.astype(str)
    logger.info(f"Read color table ({color_info.shape[0]})")

    loader = BlockIndexedLoader(args.input, args.xmin, args.xmax, args.ymin, args.ymax, args.full)
    width = int((loader.xmax - loader.xmin + 1)/args.plot_um_per_pixel)
    height= int((loader.ymax - loader.ymin + 1)/args.plot_um_per_pixel)
    logger.info(f"Image size {height} x {width}")

    # Read input file, fill the rgb matrix
    img = np.zeros((height,width,3), dtype=np.uint8)
    df = pd.DataFrame()
    keptcol = ['X','Y'] + rgb
    for df in loader:
        if df.shape[0] == 0:
            continue
        logger.info(f"Reading pixels... {df.X.iloc[-1]}, {df.Y.iloc[-1]}, {df.shape[0]}")
        df['X'] = np.clip(((df.X - loader.xmin) / args.plot_um_per_pixel).astype(int),0,width-1)
        df['Y'] = np.clip(((df.Y - loader.ymin) / args.plot_um_per_pixel).astype(int),0,height-1)
        if args.plot_top:
            for c in rgb:
                df[c] = color_info.loc[df['K1'].values, c].values
        else:
            for c in rgb:
                df[c] = 0
                for k in range(1,loader.meta['TOPK']+1):
                    df[c] += color_info.loc[df['K'+str(k)].values, c].values * df['P'+str(k)].values
        df = df.groupby(by=['X','Y']).agg({c:np.mean for c in rgb}).reset_index()
        for i,c in enumerate(rgb):
            df[c] = np.clip(np.around(df[c] * 255),0,255).astype(np.uint8)
            img[df.Y.values, df.X.values, [i]*df.shape[0]] = df[c].values

    if not args.output.endswith(".png"):
        args.output += ".png"
    cv2.imwrite(args.output,img)
    logger.info(f"Finished")



import sys, os, gzip, copy, gc, time, argparse, logging
import numpy as np
import pandas as pd
from scipy.sparse import *
import subprocess as sp
import random as rng

from ficture.utils.hexagon_fn import pixel_to_hex, hex_to_pixel

def make_dge(_args):

    parser = argparse.ArgumentParser(prog = "make_dge")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')

    parser.add_argument('--major_axis', type=str, default="Y", help='X or Y')
    parser.add_argument('--mu_scale', type=float, default=1, help='Coordinate to um translate')

    parser.add_argument('--count_header', nargs='*', type=str, default=["gn","gt", "spl","unspl","ambig"], help="Which columns correspond to UMI counts in the input")
    parser.add_argument('--group_within', nargs='*', type=str, default=[], help="Pixels are collapsed within each group. Separate by space if multiple identifier")

    parser.add_argument('--key', default = 'gt', type=str, help='gt: genetotal, gn: gene, spl: velo-spliced, unspl: velo-unspliced. Otherwise depending on customized ct_header')
    parser.add_argument('--feature_id', default = 'gene', type=str)

    parser.add_argument('--precision', type=int, default=1, help='Number of digits to store spatial location (in um), 0 for integer.')

    parser.add_argument('--n_move', type=int, default=3, help='')
    parser.add_argument('--hex_width', type=int, default=24, help='')
    parser.add_argument('--hex_radius', type=int, default=-1, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=20, help='')

    args = parser.parse_args(_args)
    logger = logging.getLogger(__name__)

    r_seed = time.time()
    rng.seed(r_seed)
    logger.info(f"Random seed {r_seed}")
    mj = args.major_axis

    # Input file and numerical columns to use as counts
    ct_header = args.count_header
    key = args.key
    if key not in ct_header:
        key = ct_header[0]
        logger.warning(f"The designated major key is not one of the specified count columns, --key is ignored the first existing key is chosen")
    if not os.path.exists(args.input):
        sys.exit(f"ERROR: cannot find input file \n {args.input}")
    with gzip.open(args.input, 'rt') as rf:
        input_header=rf.readline().strip().split('\t')
        ct_header = [v for v in input_header if v in ct_header]
        if len(ct_header) == 0:
            sys.exit("Input header does not contain the specified --count_header")
    keep_header=['X','Y','j','Group'] + ct_header
    output_header = ["random_index",'X','Y','gene'] + ct_header + args.group_within
    with open(args.output,'w') as wf:
        _=wf.write('\t'.join(output_header)+'\n')

    # basic parameters
    random_index_max=sys.maxsize//100000
    random_index_length=int(np.log10(random_index_max) ) + 1
    mu_scale = 1./args.mu_scale
    radius=args.hex_radius
    diam=args.hex_width
    n_move = args.n_move
    if n_move > diam // 2:
        n_move = diam // 4
    ovlp_buffer = diam * 2
    if radius < 0:
        radius = diam / np.sqrt(3)
    else:
        diam = int(radius*np.sqrt(3))

    adt = {x:np.sum for x in ct_header}
    dty = {x:int for x in ct_header}
    dty.update({x:str for x in ['X','Y','gene'] + args.group_within})

    n_unit = 0
    df_full = pd.DataFrame()
    for chunk in pd.read_csv(args.input, sep='\t', chunksize=1000000, dtype=dty):
        if chunk.shape[0] == 0:
            logger.info(f"Empty? Left over size {df_full.shape[0]}.")
            continue
        chunk.rename(columns={args.feature_id:'gene'}, inplace=True)
        chunk['j'] = chunk.X + '_' + chunk.Y
        chunk.X = chunk.X.astype(float)
        chunk.Y = chunk.Y.astype(float)
        ed = chunk[mj].iloc[-1]
        chunk['Group'] = ''
        if len(args.group_within) > 0:
            chunk['Group'] = chunk[args.group_within].apply(lambda x: '_'.join(x), axis=1)
        df_full = pd.concat([df_full, chunk])

        if df_full[mj].iloc[-1] - df_full[mj].iloc[0] < ovlp_buffer * args.mu_scale:
            # This chunk is too narrow, leave to process together with neighbors
            r = int(df_full[mj].iloc[-1]*args.mu_scale)
            l = int(df_full[mj].iloc[0] *args.mu_scale)
            logger.info(f"Left over size {df_full.shape[0]} ({l}, {r}).")
            continue

        left = copy.copy(df_full.loc[df_full[mj] > ed - ovlp_buffer * args.mu_scale, keep_header])

        for l in df_full.Group.unique():
            df = df_full.loc[df_full.Group.eq(l)]
            brc = df.groupby(by = ['j']).agg(adt).reset_index()
            brc = brc.merge(right = df[['j','X','Y']].drop_duplicates(subset='j'), on = 'j', how='inner')
            brc.index = range(brc.shape[0])
            brc['X'] *= mu_scale
            brc['Y'] *= mu_scale
            st = brc[mj].min()
            ed = brc[mj].max()
            pts = np.asarray(brc[['X','Y']])
            logger.info(f"Read {brc.shape[0]} pixels.")
            brc["hex_id"] = ""
            brc["random_index"] = 0
            offs_x = 0
            offs_y = 0
            while offs_x < n_move:
                while offs_y < n_move:
                    prefix  = str(offs_x)+str(offs_y)
                    x,y = pixel_to_hex(pts, radius, offs_x/n_move, offs_y/n_move)
                    hex_crd = list(zip(x,y))
                    ct = pd.DataFrame({'hex_id':hex_crd, 'tot':brc[key].values, 'X':pts[:, 0], 'Y':pts[:,1]}).groupby(by = 'hex_id').agg({'tot': sum, 'X':np.min, 'Y':np.min}).reset_index()
                    mid_ct = np.median(ct.loc[ct.tot >= args.min_ct_per_unit, 'tot'].values)
                    ct = set(ct.loc[(ct.tot >= args.min_ct_per_unit) & (ct[mj] > st + diam/2) & (ct[mj] < ed - diam/2), 'hex_id'].values)
                    if len(ct) < 2:
                        offs_y += 1
                        continue
                    hex_list = list(ct)
                    suff = [str(x[0])[-1]+str(x[1])[-1] for x in hex_list]
                    hex_dict = {x: str(rng.randint(1, random_index_max)).zfill(random_index_length) + suff[i] for i,x in enumerate(hex_list)}
                    brc["hex_id"] = hex_crd
                    brc["random_index"] = brc.hex_id.map(hex_dict)
                    sub = copy.copy(brc[brc.hex_id.isin(ct)] )

                    cnt = sub[['hex_id', 'random_index']].drop_duplicates()
                    hx = cnt.hex_id.map(lambda x : x[0])
                    hy = cnt.hex_id.map(lambda x : x[1])
                    cnt['X'], cnt['Y'] = hex_to_pixel(hx, hy, radius, offs_x/n_move, offs_y/n_move)

                    sub = sub.loc[:,['j','random_index']].merge(right = df, on='j', how = 'inner')
                    sub = sub.groupby(by = ['random_index','gene']).agg(adt).reset_index()
                    sub = sub.merge(right = cnt, on = 'random_index', how = 'inner')
                    sub['X'] = [f"{x:.{args.precision}f}" for x in sub.X.values]
                    sub['Y'] = [f"{x:.{args.precision}f}" for x in sub.Y.values]
                    sub = sub.astype({x:int for x in ct_header})
                    # Add offset combination as prefix to random_index
                    sub.random_index = prefix + sub.random_index.values
                    sub.loc[:, output_header].to_csv(args.output, mode='a', sep='\t', index=False, header=False)
                    n_unit += len(ct)
                    logger.info(f"Sliding offset {offs_x}, {offs_y}. Add {len(ct)} units, median count {mid_ct}, {n_unit} units so far.")
                    offs_y += 1
                offs_y = 0
                offs_x += 1
        df_full = copy.copy(left)
        logger.info(f"Left over size {df_full.shape[0]}.")



'''
Transform into factor space based on input model
Model contains gene names and either Dirichlet parameters (or probabilities for more general use?)
Input pixel level data will be grouped into (overlapping) hexagons
'''
import sys, os, copy, gzip, time, logging
import pickle, argparse
import numpy as np
import pandas as pd
import random as rng

from ficture.loaders.unit_loader import UnitLoader
from ficture.models.online_lda import OnlineLDA

def transform_unit(_args):

    parser = argparse.ArgumentParser(prog = "transform_unit")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output_pref', type=str, help='')
    parser.add_argument('--model', type=str, help='')

    parser.add_argument('--key', type=str, default = 'gn', help='gt: genetotal, gn: gene, spl: velo-spliced, unspl: velo-unspliced')
    parser.add_argument('--feature_id_map', type=str, default='', help='')
    parser.add_argument('--feature_id', type=str, default = 'gene', help='')
    parser.add_argument('--unit_id', type=str, default='random_index', help='')
    parser.add_argument('--unit_attr', type=str, nargs='+', default=[], help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=20, help='')

    parser.add_argument('--thread', type=int, default=1, help='')
    parser.add_argument('--b_size', type=int, default=512, help='')
    parser.add_argument('--chunksize', type=int, default=500000, help='Number of lines to read at a time')
    parser.add_argument('--debug', action='store_true', help='')

    args = parser.parse_args(_args)
    logger = logging.getLogger(__name__)
    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file.")
    if not os.path.exists(args.model):
        sys.exit("ERROR: cannot find model file.")
    if not os.path.exists(os.path.dirname(args.output_pref)):
        os.makedirs(os.path.dirname(args.output_pref))

    ### Input
    with gzip.open(args.input, 'rt') as rf:
        header=rf.readline().strip().split('\t')
    key = args.key
    unit_attr = args.unit_attr
    col_rename = {args.unit_id:'unit', args.feature_id:'gene'}
    header = [col_rename[x] if x in col_rename else x for x in header]
    usecol = [key, 'unit', 'gene'] + unit_attr
    for x in usecol:
        if x not in header:
            sys.exit(f"ERROR: {x} is not in the input file")
    adt = {key:int, 'unit':str, 'gene':str}
    adt.update({x:str for x in unit_attr})
    reader = pd.read_csv(gzip.open(args.input, 'rt'), sep='\t',\
                        chunksize=args.chunksize, skiprows=1, names=header,
                        usecols=usecol, dtype=adt)
    if args.debug:
        print("Input file reader")
        peek = pd.read_csv(gzip.open(args.input, 'rt'), sep='\t',\
                        chunksize=2, skiprows=1, names=header,
                        usecols=usecol, dtype=adt)
        print(next(peek))

    ### Model
    if args.model.endswith('.tsv.gz') or args.model.endswith('tsv'):
        model_mtx = pd.read_csv(args.model, sep='\t')
        feature_kept=list(model_mtx.gene)
        model_mtx = np.array(model_mtx.iloc[:, 1:])
        M, K = model_mtx.shape
        model = OnlineLDA(vocab=feature_kept,K=K,N=1e4,thread=args.thread,tol=1e-3)
        model.init_global_parameter(model_mtx.T)
    else:
        try:
            model = pickle.load(open(args.model, 'rb'))
            model.n_jobs = args.thread
            feature_kept = model.feature_names_in_
            K, M = model.components_.shape
            model.feature_names_in_ = None
        except:
            sys.exit("ERROR: --model must be either a tsv file containing gene name and factor-gene loadings or a pickle model object from sklearn LDA")

    factor_header = [str(x) for x in range(K)]
    ft_dict = {x:i for i,x in enumerate( feature_kept ) }
    if os.path.exists(args.feature_id_map):
        feature_name_map = {}
        with open(args.feature_id_map, 'r') as rf:
            for line in rf:
                line = line.strip().split('\t')
                feature_name_map[line[1]] = line[0]
        ft_dict = {feature_name_map[x]:i for x,i in ft_dict.items() if x in feature_name_map}
        logger.info(f"{len(ft_dict)} features from input model exist in the input data")
    logger.info(f"Model loaded with {M} features and {K} factors")



    # Unit reader
    batch_obj =  UnitLoader(reader, ft_dict, key, min_ct_per_unit=args.min_ct_per_unit, unit_attr=unit_attr)

    # Transform
    post_count = np.zeros((K, M))
    n_unit = 0
    n_batch= 0
    oheader = [args.unit_id, key] + unit_attr + ["topK", "topP"] + factor_header
    out_f = args.output_pref + ".fit_result.tsv.gz"
    with gzip.open(out_f, 'wt') as wf:
        wf.write('\t'.join(oheader) + '\n')
    t0 = time.time()
    while batch_obj.update_batch(args.b_size):
        if args.debug:
            print(type(batch_obj.mtx))
            print(batch_obj.mtx.shape)
        theta = model.transform(batch_obj.mtx)
        post_count += np.array(theta.T @ batch_obj.mtx)
        n_batch += 1
        n_unit  += theta.shape[0]
        t1 = time.time() - t0
        if args.debug:
            print(batch_obj.brc[:2])
        batch_obj.brc['topK'] = np.argmax(theta, axis = 1)
        batch_obj.brc['topP'] = theta.max(axis = 1)
        batch_obj.brc = pd.concat([batch_obj.brc, pd.DataFrame(theta, columns=factor_header)], axis=1)
        batch_obj.brc.rename(columns = {'unit':args.unit_id}, inplace=True)
        batch_obj.brc[oheader].to_csv(out_f, sep='\t', index=False, header=False, float_format='%.3e', mode='a', compression={"method":"gzip"})
        logger.info(f"Transformed {n_batch} batches with total {n_unit} units, {t1/60:2f}min")
        if args.debug:
            break

    if args.debug:
        sys.exit()

    out_f = args.output_pref + ".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({'gene': feature_kept}),\
            pd.DataFrame(post_count.T, dtype='float64',\
                            columns = [str(k) for k in range(K)])],\
                axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})



import sys, os, warnings, copy
import numpy as np
import pandas as pd
import sklearn.neighbors
import sklearn.mixture

from ficture.utils.hexagon_fn import pixel_to_hex, hex_to_pixel


def filter_by_density_mixture(df, key, radius, n_move, args):
    pt = pd.DataFrame()
    m0v=[]
    m1v=[]
    hex_area = radius**2 * np.sqrt(3) * 3 / 2
    for i in range(n_move):
        for j in range(n_move):
            cnt = pd.DataFrame()
            cnt["hex_x"], cnt["hex_y"] = pixel_to_hex(np.asarray(df.loc[:, ['X','Y']]), radius, i/n_move, j/n_move)
            cnt[key] = copy.copy(df[key].values)
            cnt = cnt.groupby(by = ['hex_x','hex_y']).agg({key:sum}).reset_index()
            cnt = cnt[cnt[key] > hex_area * args.min_abs_mol_density_squm]
            if cnt.shape[0] < 10:
                continue
            if args.hard_threshold > 0:
                cnt['det'] = cnt[key] > hex_area * args.hard_threshold
            else:
                v = np.log10(cnt[key].values).reshape(-1, 1)
                print(f"{len(v)}")
                if len(v) > args.max_npts_to_fit_model:
                    indx = np.random.choice(len(v), int(args.max_npts_to_fit_model), replace=False)
                    gm = sklearn.mixture.GaussianMixture(n_components=2, random_state=0).fit(v[indx])
                else:
                    gm = sklearn.mixture.GaussianMixture(n_components=2, random_state=0).fit(v)
                lab_keep = np.argmax(gm.means_.squeeze())
                cnt['det'] = gm.predict(v) == lab_keep
                m0=(10**gm.means_.squeeze()[lab_keep])/hex_area
                m1=(10**gm.means_.squeeze()[1-lab_keep])/hex_area
                if m1 > m0 * 0.5 or m0 < args.min_abs_mol_density_squm:
                    v = cnt[key].values.reshape(-1, 1)
                    if len(v) > args.max_npts_to_fit_model:
                        indx = np.random.choice(len(v), int(args.max_npts_to_fit_model), replace=False)
                        gm = sklearn.mixture.GaussianMixture(n_components=2, random_state=0).fit(v[indx])
                    else:
                        gm = sklearn.mixture.GaussianMixture(n_components=2, random_state=0).fit(v)
                    lab_keep = np.argmax(gm.means_.squeeze())
                    cnt['det'] = gm.predict(v) == lab_keep
                    m0=gm.means_.squeeze()[lab_keep]/hex_area
                    m1=gm.means_.squeeze()[1-lab_keep]/hex_area
                if m0 < args.min_abs_mol_density_squm:
                    continue
                if m1 > m0 * .7:
                    cnt['det'] = 1
                m0v.append(m0)
                m1v.append(m1)
            if cnt.det.eq(True).sum() < 2:
                continue
            m0 = cnt.loc[cnt.det.eq(True), key].median()/hex_area
            m1 = cnt.loc[cnt.det.eq(False), key].median()/hex_area
            m0v.append(m0)
            m1v.append(m1)
            anchor_x, anchor_y = hex_to_pixel(cnt.loc[cnt.det.eq(True), 'hex_x'].values, cnt.loc[cnt.det.eq(True), 'hex_y'].values, radius,i/n_move,j/n_move)
            pt = pd.concat([pt,\
                    pd.DataFrame({'x':anchor_x, 'y':anchor_y})])
    return pt, np.mean(m0v), np.mean(m1v)



import sys, os, copy, gzip, logging, re
import numpy as np
import pandas as pd
from scipy.sparse import coo_array, csr_array, vstack
import sklearn.neighbors
from sklearn.preprocessing import normalize

from ficture.utils import utilt

class factor_space_stream:
    """
    Load spatial factor decoding results by chunk
    Assuming both input and queries are sorted by index_axis
    Answer queries of spatial locations by averaging over the nearest neighbors
    """
    def __init__(self, file, index_axis, debug =0, init_pos = 0,\
                 sparsify = 1e-4, buffer = 30, chunksize = 100000,\
                 x = "X", y = "Y", factor_header = None) -> None:
        assert os.path.exists(file), "File does not exist: {}".format(file)
        assert index_axis in [x, y], "Index axis must be the same as either the provided x (X) or y (Y)[]"
        self.debug = debug
        self.file = file
        self.index_col = index_axis
        self.index_axis = 0 if index_axis == x else 1
        self.sparsify = sparsify
        self.buffer = buffer
        self.factor_header = factor_header
        if self.factor_header is None:
            with gzip.open(self.file, 'rt') as rf:
                input_header = rf.readline().strip().split('\t')
            self.factor_header = utilt.get_string_with_integer_suff(input_header)
        self.K = len(self.factor_header)
        self.reader = pd.read_csv(gzip.open(self.file, 'rt'), sep='\t', \
                                  usecols = [x,y] + self.factor_header, \
                                  chunksize=chunksize)
        self.recolumn = {x:"X", y:"Y"}
        self.recolumn.update({x:str(k) for k,x in enumerate(self.factor_header)})
        self.factor_header = [str(k) for k in range(self.K)]
        self.pts = np.empty((0, 2))
        self.factor_loading = csr_array((0, self.K))
        self.current_ymax = -1
        self.current_ymin = -1
        self.file_is_open = True
        self.update_reference(init_pos, init_pos)
        if self.debug:
            print(f"Current view is from {self.current_ymin} to {self.current_ymax}")

    def read_chunk(self, ymin):
        try:
            chunk = next(self.reader)
            chunk.drop_duplicates(subset = ['X','Y'], inplace = True)
        except StopIteration:
            return 0
        self.current_ymax = chunk[self.index_col].iloc[-1]
        chunk = chunk.loc[chunk[self.index_col] > ymin, :]
        if chunk.shape[0] == 0:
            return 1
        chunk.rename(columns=self.recolumn, inplace=True)
        self.pts = np.vstack((self.pts, chunk.loc[:, ["X","Y"]].values))
        chunk = np.array(chunk.loc[:, self.factor_header].values)
        chunk[chunk < self.sparsify] = 0
        chunk = csr_array(chunk)
        chunk.eliminate_zeros()
        self.factor_loading = vstack((self.factor_loading, chunk))
        self.current_ymax = self.pts[-1, self.index_axis]
        return 1

    def update_reference(self, ymax, ymin = None):
        if ymin is None:
            ymin = self.current_ymax
        ymin -= self.buffer
        ymax += self.buffer
        indx_buffer = self.pts[:, self.index_axis] > ymin
        self.pts  = self.pts[indx_buffer, :]
        self.factor_loading = self.factor_loading[indx_buffer, :]
        while self.current_ymax < ymax and self.file_is_open:
            if self.debug > 1:
                print(f"Current cursor: {self.current_ymax}, target: {ymax}")
            self.file_is_open = self.read_chunk(ymin)
        self.ref = sklearn.neighbors.BallTree(self.pts)
        self.current_ymax = self.pts[:, self.index_axis].max()
        self.current_ymin = self.pts[:, self.index_axis].min()

    def impute_factor_loading(self, pos, k = 1, radius = 5, halflife = .7, include_self = False):
        ymin = pos[:, self.index_axis].min()
        ymax = pos[:, self.index_axis].max()
        if ymax > self.current_ymax:
            self.update_reference(ymax, ymin)
        nu = np.log(.5) / np.log(halflife)
        dist, indx = self.ref.query(pos, k = k, return_distance = True, sort_results = True)
        min_dist = 0
        if include_self:
            min_dist = -1
        mask = (dist > min_dist) & (dist < radius)
        indx = [[x for k,x in enumerate(v) if mask[i,k]] for i,v in enumerate(indx) ]
        dist = 1. - (dist / radius) ** nu
        dist[~mask] = -1
        r = [i for i,y in enumerate(indx) for x in range(len(y))]
        c = [x for y in indx for x in y]
        w = [x for y in dist for x in y if x >= 0]
        mtx = coo_array((w, (r, c)), shape = (pos.shape[0], self.pts.shape[0])).tocsr()
        mtx = normalize(mtx, norm='l1', axis=1, copy=False)
        mtx.eliminate_zeros()
        return mtx @ self.factor_loading





class factor_space:
    """
    Load spatial factor decoding results
    Answer queries of spatial locations by averaging over the nearest neighbors
    """
    def __init__(self, file, debug = 0, sparsify = 1e-4,\
                 x = "x", y = "y", factor_header = None) -> None:
        assert os.path.exists(file), "File does not exist: {}".format(file)
        self.file = file
        self.debug = debug
        self.sparsify = sparsify
        self.factor_header = factor_header
        with gzip.open(self.file, 'rt') as rf:
            input_header = rf.readline().strip().split('\t')
        if self.factor_header is None:
            self.factor_header = utilt.get_string_with_integer_suff(input_header)
        self.K = len(self.factor_header)
        self.reader = pd.read_csv(gzip.open(self.file, 'rt'), sep='\t', \
                                  chunksize=100000)
        self.recolumn = {x.lower():"X", y.lower():"Y"}
        self.recolumn.update({x.lower():str(k) for k,x in enumerate(self.factor_header)})
        self.factor_header = [str(k) for k in range(self.K)]
        self.pts = np.empty((0, 2))
        self.factor_loading = csr_array((0, self.K))
        self.update_reference()

    def read_chunk(self):
        try:
            chunk = next(self.reader)
            chunk.columns = [x.lower() for x in chunk.columns]
        except StopIteration:
            return 0
        chunk.rename(columns=self.recolumn, inplace=True)
        self.pts = np.vstack((self.pts, chunk.loc[:, ["X","Y"]].values))
        chunk = np.array(chunk.loc[:, self.factor_header].values)
        chunk[chunk < self.sparsify] = 0
        chunk = csr_array(chunk)
        chunk.eliminate_zeros()
        self.factor_loading = vstack((self.factor_loading, chunk))
        if self.debug > 1:
            print(f"Read chunk {chunk.shape}, {self.factor_loading.shape}")
        return 1

    def update_reference(self):
        while True:
            if self.read_chunk() == 0:
                break
        self.ref = sklearn.neighbors.BallTree(self.pts)
        if self.debug:
            print(f"Build reference with {self.pts.shape[0]} points")

    def impute_factor_loading(self, pos, k = 1, radius = 5, halflife = .7, include_self = False):
        nu = np.log(.5) / np.log(halflife)
        dist, indx = self.ref.query(pos, k = k, return_distance = True, sort_results = True)
        min_dist = 0
        if include_self:
            min_dist = -1
        if self.debug > 1:
            v = np.min(dist,axis=1)
            print(radius, sum(v > radius), len(v))
        mask = (dist > min_dist) & (dist < radius)
        indx = [[x for k,x in enumerate(v) if mask[i,k]] for i,v in enumerate(indx) ]
        dist = 1. - (dist / radius) ** nu
        dist[~mask] = -1
        r = [i for i,y in enumerate(indx) for x in range(len(y))]
        c = [x for y in indx for x in y]
        w = [x for y in dist for x in y if x >= 0]
        mtx = coo_array((w, (r, c)), shape = (pos.shape[0], self.pts.shape[0])).tocsr()
        mtx = normalize(mtx, norm='l1', axis=1, copy=False)
        mtx.eliminate_zeros()
        return mtx @ self.factor_loading





class StreamUnit:
    """
    Read a long format DGE by chunk
    Output matrices sequentially
    """
    def __init__(self, reader, unit_id, key, ft_dict, feature_id="gene",
                 min_batch_size = 256, min_ct_per_unit = 2):
        self.reader = reader
        self._unit_id = unit_id
        self._feature_id = feature_id
        self._key = key
        self._min_n = min_batch_size
        self._min_c = min_ct_per_unit
        self.ft_dict = copy.copy(ft_dict)
        self.M = len(self.ft_dict)
        self.df = pd.DataFrame()

    def get_matrix(self):
        for chunk in self.reader:
            chunk = chunk[chunk.gene.isin(self.ft_dict)]
            chunk.rename(inplace=True, \
                         columns = {self._unit_id:'j', self._feature_id:'gene'})
            # Incomplete left over
            last_indx = chunk.j.iloc[-1]
            left = copy.copy(chunk[chunk.j.eq(last_indx)])
            chunk = chunk.loc[~chunk.j.eq(last_indx), :]

            ct = chunk.groupby(by = ['j']).agg({self._key: sum}).reset_index()
            kept_unit = set(ct.loc[ct[self._key] > self._min_c, "j"].values)
            self.df = pd.concat([self.df, chunk[chunk.j.isin(kept_unit)]])
            if len(self.df.j.unique()) < self._min_n: # Wait until more data
                self.df = pd.concat([self.df, left])
                continue

            # Total molecule count per unit
            brc = self.df.groupby(by = ['j']).agg({self._key: sum}).reset_index()
            brc = brc[brc[self._key] > self._min_c]
            brc.index = range(brc.shape[0])
            self.df = self.df[self.df.j.isin(brc.j.values)]
            # Make DGE
            barcode_kept = list(brc.j.values)
            bc_dict = {x:i for i,x in enumerate( barcode_kept ) }
            indx_row = [ bc_dict[x] for x in self.df['j']]
            indx_col = [ self.ft_dict[x] for x in self.df['gene']]
            N = len(barcode_kept)
            mtx = coo_array((self.df[self._key].values, (indx_row, indx_col)), shape=(N, self.M)).tocsr()
            yield mtx
            self.df = copy.copy(left)

        if len(self.df.j.unique()) > 1:
            brc = self.df.groupby(by = ['j']).agg({self._key: sum}).reset_index()
            brc = brc[brc[self._key] > self._min_c]
            brc.index = range(brc.shape[0])
            self.df = self.df[self.df.j.isin(brc.j.values)]
            # Make DGE
            barcode_kept = list(brc.j.values)
            bc_dict = {x:i for i,x in enumerate( barcode_kept ) }
            indx_row = [ bc_dict[x] for x in self.df['j']]
            indx_col = [ self.ft_dict[x] for x in self.df['gene']]
            N = len(barcode_kept)
            mtx = coo_array((self.df[self._key].values, (indx_row, indx_col)), shape=(N, self.M)).tocsr()
            yield mtx



### Filtering pixeld based on input boundaries (in the form of polygons)

import sys, os, gzip, argparse, warnings, logging
import numpy as np
import pandas as pd
from matplotlib.path import Path
from geopandas import GeoSeries
import shapely, geojson
from shapely.geometry import Point, Polygon

from ficture.utils.utilt import extract_polygons_from_json, svg_parse_list

def filter_by_boundary(_args):

    parser = argparse.ArgumentParser(prog = "filter_by_boundary")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--boundary', type=str, help='Boundary info (currently only support geojson or tsv with a path column in svg format)')
    parser.add_argument('--boundary_unit_in_um', type=float, default=1, help='')
    parser.add_argument('--feature', type=str, default='', help='')
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--offset_x', type=float, default=0, help='In um')
    parser.add_argument('--offset_y', type=float, default=0, help='In um')
    parser.add_argument('--transpose', action='store_true')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    if os.path.exists(args.output):
        warnings.warn("Output file already exists")
    if not os.path.exists(args.input):
        sys.exit("Cannot find input file")
    if not os.path.exists(args.boundary):
        sys.exit("Cannot find input file")

    # logging.basicConfig(level= getattr(logging, "INFO", None))
    logger = logging.getLogger(__name__)

    if args.boundary.endswith('.geojson'):
        vertices = extract_polygons_from_json(args.boundary)
        vertices = [(x * args.boundary_unit_in_um + np.array([args.offset_x, args.offset_y])) for x in vertices]
        n_poly = len(vertices)
        poly_list = [shapely.polygons(x) for x in vertices]
        poly_list = [x if x.is_valid else x.buffer(0) for x in poly_list]
        poly = shapely.unary_union(poly_list)
        # poly = shapely.unary_union([shapely.polygons(x) for x in vertices])
    elif args.boundary.endswith('.tsv'):
        poly_df = pd.read_csv(args.boundary, sep='\t')
        paths = poly_df.path.tolist()
        poly_list = []
        for i in range(len(paths)):
            path = eval(paths[i] )
            codes, verts = svg_parse_list(path)
            verts = np.array(verts)
            # Translate into pixel coordinate
            verts *= args.boundary_unit_in_um
            verts[:, 0] += args.offset_x
            verts[:, 1] += args.offset_y
            mpl_path = Path(verts, codes)
            poly_list.append(Polygon(mpl_path.to_polygons()[0]))
            print(verts.min(axis=0), verts.max(axis=0))
        n_poly = len(poly_list)
        poly_list = [x if x.is_valid else x.buffer(0) for x in poly_list]
        poly = shapely.unary_union(poly_list)
    else:
        sys.exit("Unknown boundary format")

    logger.info(f"Read boundary info with {n_poly} polygons, total area {poly.area:.1f} um^2")

    gene_kept = set()
    if os.path.exists(args.feature):
        feature = pd.read_csv(args.feature, sep='\t', header=0)
        gene_kept = set(feature.gene.values)
        logger.info(f"Read feature info with {len(gene_kept)} genes")

    with gzip.open(args.input, 'rt') as rf:
        header = rf.readline()
    if args.output.endswith('.gz'):
        with gzip.open(args.output, 'wt') as wf:
            _=wf.write(header)
    else:
        with open(args.output, 'w') as wf:
            _=wf.write(header)

    chunk_size = 500000
    for chunk in pd.read_csv(gzip.open(args.input, 'rb'),\
        sep='\t', header=0, chunksize=chunk_size):
        chunk.rename(columns={'x': 'X', 'y': 'Y'}, inplace=True)
        if len(gene_kept) != 0:
            chunk = chunk.loc[chunk.gene.isin(gene_kept)]
            if chunk.shape[0] == 0:
                continue
        print(chunk.loc[:, ['X', 'Y']].max(axis = 0).values / args.mu_scale)
        print(chunk.loc[:, ['X', 'Y']].min(axis = 0).values / args.mu_scale)
        if args.transpose:
            pts = GeoSeries(map(Point, zip(chunk.Y.values / args.mu_scale,\
                                        chunk.X.values / args.mu_scale)))
        else:
            pts = GeoSeries(map(Point, zip(chunk.X.values / args.mu_scale,\
                                        chunk.Y.values / args.mu_scale)))
        chunk = chunk.loc[pts.within(poly).values, :]
        logger.info(f"Output {chunk.shape[0]} rows ...")
        if chunk.shape[0] == 0:
            continue
        chunk.to_csv(args.output, mode='a', sep='\t', index=False, header=False)



### Read a batched pixel file, construct slda minibatch

import sys, os, gzip, copy, re
import numpy as np
import pandas as pd

from scipy.sparse import coo_array
import sklearn.neighbors
import sklearn.preprocessing
from joblib.parallel import Parallel, delayed

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
from ficture.utils import utilt
from ficture.models.slda_minibatch import minibatch

class PixelMinibatch:

    def __init__(self, reader, ft_dict, batch_id, key, mu_scale, radius, halflife, adj_penal=-1, precision=0.1, thread=1, verbose=0) -> None:
        self.df_full = pd.DataFrame()
        self.pixel_reader = reader
        self.batch_id = batch_id
        self.mu_scale = mu_scale
        self.file_is_open = True
        self.ft_dict = ft_dict
        self.M = len(self.ft_dict)
        self.key = key
        self.precision = np.clip(precision, 0.1, np.inf)
        self.radius = radius
        self.adj_penal = adj_penal
        self.nu = np.log(.5) / np.log(halflife)
        self.out_buff = self.radius * halflife
        self.thread = thread
        self.verbose = verbose

    def load_anchor(self, anchor_file, anchor_in_um = True):
        self.grid_info = pd.read_csv(anchor_file,sep='\t')
        header_map = {"X":"x", "Y":"y"}
        self.factor_header = []
        for x in self.grid_info.columns:
            y = re.match('^[A-Za-z]*_*(\d+)$', x)
            if y:
                header_map[y.group(0)] = y.group(1)
                self.factor_header.append(int(y.group(1)) )
        self.grid_info.rename(columns = header_map, inplace=True)
        if not anchor_in_um:
            self.grid_info.x *= self.mu_scale
            self.grid_info.y *= self.mu_scale
        self.n = self.grid_info.shape[0]
        self.grid_info.index = range(self.n)
        self.ref = sklearn.neighbors.BallTree(np.array(self.grid_info.loc[:, ['x','y']]))
        self.factor_header.sort()
        self.factor_header = [str(x) for x in self.factor_header]
        self.K = len(self.factor_header)
        self.adj_mtx = None

    def _anchor_adj(self):
        if self.adj_penal < 0:
            self.adj_penal = self.radius * 2
        self.adj_mtx = sklearn.neighbors.radius_neighbors_graph(self.ref, self.adj_penal, mode='connectivity', include_self=True) # Do we want to include diagonal?

    def read_chunk(self, nbatch):
        batch_ids = set()
        while len(batch_ids) <= nbatch:
            try:
                chunk = next(self.pixel_reader)
            except StopIteration:
                self.file_is_open = False
                break
            chunk = chunk.loc[chunk.gene.isin(self.ft_dict) & \
                              (chunk[self.key] > 0), :]
            chunk.X = (chunk.X * self.mu_scale / self.precision).astype(int)
            chunk.Y = (chunk.Y * self.mu_scale / self.precision).astype(int)
            chunk = chunk.groupby(by=[self.batch_id,"gene","X","Y"]).agg({self.key:np.sum}).reset_index()
            random_pref = chunk[self.batch_id].map(lambda x : x[-5:]).values
            chunk['j'] = random_pref + '_' + chunk.X.astype(str) + '_' + chunk.Y.astype(str)
            chunk.X *= self.precision
            chunk.Y *= self.precision
            # Keep pixels close enough to at least one anchor
            pts = chunk[["j", "X", "Y"]].drop_duplicates(subset="j")
            dist, indx = self.ref.query(X = np.array(pts[['X','Y']]), k = 1, return_distance = True)
            dist = dist.squeeze()

            kept_pixel = dist < self.radius
            if np.sum(kept_pixel) == 0:
                continue
            chunk = chunk.loc[chunk.j.isin(pts.loc[kept_pixel, 'j'].values), :]
            batch_ids.update(set(chunk[self.batch_id].unique()))
            self.df_full = pd.concat([self.df_full, chunk], axis=0)
            # print(f"Read {len(batch_ids)} minibatches")

        left = pd.DataFrame()
        if self.file_is_open:
            last_indx = self.df_full[self.batch_id].iloc[-1]
            left = copy.copy(self.df_full.loc[self.df_full[self.batch_id].eq(last_indx), :])
            self.df_full = self.df_full.loc[~self.df_full[self.batch_id].eq(last_indx), :]

        ### Process chunk of data
        self.batch_index = list(self.df_full[self.batch_id].unique() )
        self.brc = self.df_full[[self.batch_id,"j","X","Y"]].drop_duplicates(subset=["j"])
        self.N0 = self.brc.shape[0]
        self.brc.index = range(self.N0)
        self.df_full = self.df_full.groupby(by = [self.batch_id, "j", "gene"]).agg({self.key:sum}).reset_index()
        print(f"Read {self.N0} pixels, forming {len(self.batch_index)} batches.")
        self.bt = sklearn.neighbors.BallTree(np.asarray(self.brc[['X','Y']]))
        # Make DGE
        barcode_kept = list(self.brc['j'])
        bc_dict = {x:i for i,x in enumerate( barcode_kept ) }
        indx_row = [ bc_dict[x] for x in self.df_full['j']]
        indx_col = [ self.ft_dict[x] for x in self.df_full['gene']]
        self.dge_mtx = coo_array((self.df_full[self.key], (indx_row, indx_col)), shape=(self.N0, self.M)).tocsr()
        self.df_full = left
        return len(self.batch_index)

    def _prepare_batch(self, b, init_bound):

        indx = self.brc[self.batch_id].eq(b)
        x_min, x_max = self.brc.loc[indx, 'X'].min(), self.brc.loc[indx, 'X'].max()
        y_min, y_max = self.brc.loc[indx, 'Y'].min(), self.brc.loc[indx, 'Y'].max()
        grid_indx = (self.grid_info.x >= x_min - self.radius) &\
                    (self.grid_info.x <= x_max + self.radius) &\
                    (self.grid_info.y >= y_min - self.radius) &\
                    (self.grid_info.y <= y_max + self.radius)
        grid_pt = self.grid_info.loc[grid_indx, ["x","y"]]
        if grid_pt.shape[0] < 10:
            return None, None, None, None

        # Initilize anchor
        theta = np.array(self.grid_info.loc[grid_indx, self.factor_header])
        theta = sklearn.preprocessing.normalize(np.clip(theta, init_bound, 1.-init_bound), norm='l1', axis=1)
        n = theta.shape[0]

        # Pixel to anchor weight
        indx, dist = self.bt.query_radius(X = np.array(grid_pt), r = self.radius, return_distance = True)
        r_indx = [i for i,x in enumerate(indx) for y in range(len(x))]
        c_indx = [x for y in indx for x in y]
        wij = np.array([x for y in dist for x in y])
        wij = 1-(wij / self.radius)**self.nu
        wij = coo_array((wij, (r_indx,c_indx)),shape=(n,self.N0)).tocsc().T
        wij.eliminate_zeros()
        nchoice=(wij != 0).sum(axis = 1)
        b_indx = np.arange(self.N0)[nchoice > 0]
        wij = wij[b_indx, :]
        wij.data = np.clip(wij.data, .05, .95)
        return b_indx, grid_pt, wij, theta

    def one_batch(self, batch_index, slda, init_bound):

        pixel_result = pd.DataFrame()
        anchor_result = pd.DataFrame()
        post_count = np.zeros((self.K, self.M))
        for b in batch_index:
            b_indx, grid_pt, wij, theta = self._prepare_batch(b, init_bound)
            if b_indx is None:
                continue
            N = len(b_indx)
            grid_pt = np.array(grid_pt)
            psi_org = sklearn.preprocessing.normalize(wij, norm='l1', axis=1)
            batch = minibatch()
            batch.init_from_matrix(self.dge_mtx[b_indx, :], grid_pt, wij, psi = psi_org, m_gamma = theta)
            _ = slda.do_e_step(batch)

            tmp = self.brc.iloc[b_indx, :]
            x_min, x_max = tmp.X.min(), tmp.X.max()
            y_min, y_max = tmp.Y.min(), tmp.Y.max()
            v = np.arange(N)[(tmp.X > x_min+self.out_buff) &\
                             (tmp.X < x_max-self.out_buff) &\
                             (tmp.Y > y_min+self.out_buff) &\
                             (tmp.Y < y_max-self.out_buff)]
            post_count += batch.phi[v, :].T @ batch.mtx[v, :]

            tmp = copy.copy(self.brc.loc[b_indx, ['j','X','Y']] )
            tmp.index = range(tmp.shape[0])
            tmp = pd.concat([tmp, pd.DataFrame(batch.phi, \
                             columns = self.factor_header)], axis = 1)
            tmp = tmp.iloc[v, :]
            pixel_result = pd.concat([pixel_result, tmp], axis = 0)

            expElog_theta = np.exp(utilt.dirichlet_expectation(batch.gamma))
            expElog_theta/= expElog_theta.sum(axis = 1).reshape((-1, 1))
            tmp = pd.DataFrame({'minibatch':b,'X':grid_pt[:,0],'Y':grid_pt[:,1]})
            asum = batch.psi.T @ batch.mtx.sum(axis = 1).reshape((-1, 1))
            tmp['avg_size'] = np.array(asum).reshape(-1)
            for v in range(self.K):
                tmp[str(v)] = expElog_theta[:, v]
            tmp = tmp.loc[(tmp.X > x_min+self.out_buff) & \
                          (tmp.X < x_max-self.out_buff) & \
                          (tmp.Y > y_min+self.out_buff) & \
                          (tmp.Y < y_max-self.out_buff), :]
            anchor_result = pd.concat([anchor_result, tmp], axis = 0)
        return post_count, pixel_result, anchor_result

    def run_chunk(self, slda, init_bound):
        if self.thread > 1:
            idx_slices = [[ self.batch_index[x] for x in y ] for y in utilt.gen_even_slices(len(self.batch_index), self.thread)]
            with Parallel( n_jobs=self.thread, backend='threading', verbose=self.verbose) as parallel:
                result_list = parallel(delayed(self.one_batch)(idx, slda, init_bound) for idx in idx_slices)
            pixel_result = pd.DataFrame()
            anchor_result = pd.DataFrame()
            post_count = np.zeros((self.K, self.M))
            for obj in result_list:
                post_count += obj[0]
                pixel_result = pd.concat([pixel_result, obj[1]], axis = 0)
                anchor_result = pd.concat([anchor_result, obj[2]], axis = 0)
            return post_count, pixel_result, anchor_result
        else:
            return self.one_batch(self.batch_index, slda, init_bound)



    def run_chunk_penalized(self, slda, init_bound):
        assert slda._zeta > 0 and slda._zeta < 1, "To run slda with penalized likelihood, please set slda.zeta within (0,1)"
        if self.adj_mtx is None:
            self._anchor_adj()
        pixel_result = pd.DataFrame()
        anchor_result = pd.DataFrame()
        post_count = np.zeros((self.K, self.M))
        for b in self.batch_index:
            b_indx, grid_pt, wij, theta = self._prepare_batch(b, init_bound)
            if b_indx is None:
                continue
            N = len(b_indx)
            anchor_index = grid_pt.index.to_list()
            grid_pt = np.array(grid_pt)
            psi_org = sklearn.preprocessing.normalize(wij, norm='l1', axis=1)
            batch = minibatch()
            batch.init_from_matrix(self.dge_mtx[b_indx, :], grid_pt, wij, psi = psi_org, m_gamma = theta, anchor_adj = self.adj_mtx[anchor_index, :][:, anchor_index])

            _ = slda.update_lambda_penalized(batch)

            tmp = self.brc.iloc[b_indx, :]
            x_min, x_max = tmp.X.min(), tmp.X.max()
            y_min, y_max = tmp.Y.min(), tmp.Y.max()
            v = np.arange(N)[(tmp.X > x_min+self.out_buff) &\
                             (tmp.X < x_max-self.out_buff) &\
                             (tmp.Y > y_min+self.out_buff) &\
                             (tmp.Y < y_max-self.out_buff)]
            post_count += batch.phi[v, :].T @ batch.mtx[v, :]

            tmp = copy.copy(self.brc.loc[b_indx, ['j','X','Y']] )
            tmp.index = range(tmp.shape[0])
            tmp = pd.concat([tmp, pd.DataFrame(batch.phi, \
                             columns = self.factor_header)], axis = 1)
            tmp = tmp.iloc[v, :]
            pixel_result = pd.concat([pixel_result, tmp], axis = 0)

            expElog_theta = np.exp(utilt.dirichlet_expectation(batch.gamma))
            expElog_theta/= expElog_theta.sum(axis = 1).reshape((-1, 1))
            tmp = pd.DataFrame({'minibatch':b,'X':grid_pt[:,0],'Y':grid_pt[:,1]})
            asum = batch.psi.T @ batch.mtx.sum(axis = 1).reshape((-1, 1))
            tmp['avg_size'] = np.array(asum).reshape(-1)
            for v in range(self.K):
                tmp[str(v)] = expElog_theta[:, v]
            tmp = tmp.loc[(tmp.X > x_min+self.out_buff) & \
                          (tmp.X < x_max-self.out_buff) & \
                          (tmp.Y > y_min+self.out_buff) & \
                          (tmp.Y < y_max-self.out_buff), :]
            anchor_result = pd.concat([anchor_result, tmp], axis = 0)
        return post_count, pixel_result, anchor_result



import sys, io, os, gzip, glob, copy, re, time, pickle, warnings, importlib
import numpy as np
import pandas as pd
import base64

from sklearn.preprocessing import normalize
import scipy.cluster
import scipy.stats
import scipy.spatial
import scipy.linalg

import ete3
from ete3 import Tree, TreeStyle, NodeStyle
os.environ['QT_QPA_PLATFORM']='offscreen'

def logrank(x):
    v = scipy.stats.rankdata(x)
    return - np.log( 1-(v-.5)/len(v) )

def cor_logrank(orgmtx):
    K, M = orgmtx.shape
    rankmtx = np.zeros((K, M), dtype=int)
    for k in range(K):
        rankmtx[k, :] = scipy.stats.rankdata(orgmtx[k, :])
    corlogrank = np.corrcoef(-np.log( 1-(rankmtx-.5)/M ) )
    return corlogrank

def NJ_logrank(orgmtx):
    K, M = orgmtx.shape
    rankmtx = np.zeros((K, M), dtype=int)
    for k in range(K):
        rankmtx[k, :] = scipy.stats.rankdata(orgmtx[k, :])
    corlogrank = np.corrcoef(-np.log( 1-(rankmtx-.5)/M ) )
    node = {} # node id -> [leaves, profile]
    active_node = []
    for k in range(K):
        node[k] = [[k], orgmtx[k, :]]
        active_node.append(k)
    pairwise_dict = {} # (id1, id2) -> score
    for k in range(K-1):
        for l in range(k+1, K):
            pairwise_dict[(k, l)] = corlogrank[k, l]
    Z = []
    for s in range(K-1):
        candi = [] # [(id1, id2), score]
        for i,v1 in enumerate(active_node[:-1]):
            for j in range(i+1, len(active_node)):
                v2 = active_node[j]
                if (v1, v2) not in pairwise_dict:
                    vec1 = logrank(node[v1][1])
                    vec2 = logrank(node[v2][1])
                    score = np.corrcoef(vec1, vec2)[0,1]
                    candi.append([(v1, v2), score])
                    pairwise_dict[ (v1, v2) ] = score
                else:
                    score = pairwise_dict[(v1, v2)]
                    candi.append([(v1, v2), score])
        if len(candi) == 0:
            break
        candi.sort(key = lambda x : x[1], reverse=True)
        v1, v2 = candi[0][0]
        lvs = node[v1][0]+node[v2][0]
        node[K+s] = [lvs, orgmtx[lvs, :].mean(axis = 0)]
        active_node.remove(v1)
        active_node.remove(v2)
        active_node.append(K+s)
        Z.append([ v1, v2, 1-candi[0][1], len(lvs) ])
    return Z

def visual_hc(model_prob, weight, top_gene, node_color=None, circle=False, vertical=False, output_f=None, cprob_cut=.99):

    K = model_prob.shape[0]
    assert len(weight) == K, "model_prob.shape[0] != len(weight)"
    assert len(top_gene) == K, "len(top_gene) != K"

    model_prob = normalize(np.array(model_prob), norm='l1', axis=1)
    weight = np.array(weight)
    weight /= weight.sum()
    weight_anno = ["%.2e" % x if x < 0.1 else "%.3f" % x for x in weight]
    v = np.argsort(weight)[::-1]
    w = np.cumsum(weight[v] )
    if sum(w > cprob_cut) == 0:
        k = K - 1
    else:
        k = np.arange(K)[w > cprob_cut][0]
    kept_factor = v[:(k+1)].astype(str)

    # # Hierarchical clustering
    # cd_dist = scipy.spatial.distance.pdist(model_prob, metric='cosine')
    # Z_hc = scipy.cluster.hierarchy.linkage(cd_dist, method="complete")
    # Z_hc = NJ_logrank(model_prob)

    corlogrank = cor_logrank(model_prob)
    corlogrank = np.nan_to_num(corlogrank, copy=False)
    cd_dist = 1 - .5 * (corlogrank + corlogrank.T)
    np.fill_diagonal(cd_dist, 0)
    cd_dist = scipy.spatial.distance.squareform(cd_dist)
    Z_hc = scipy.cluster.hierarchy.linkage(cd_dist, method="complete")

    # Construct tree object from the clustering
    R, T = scipy.cluster.hierarchy.to_tree(Z_hc, rd=True)
    tr = Tree()
    tr.dist=0
    tr.name='root'
    node_dict = {R.id:tr}
    stack = [R]
    while stack:
        node = stack.pop()
        c_dist = node.dist / 2
        for c in [node.left, node.right]:
            if c:
                ch = Tree()
                ch.dist = c_dist
                ch.name = str(c.id)
                _=node_dict[node.id].add_child(ch)
                node_dict[c.id] = ch
                stack.append(c)

    node_list = [x for x in tr.traverse() if x.is_leaf() and x.name in kept_factor]
    subtr = tr.copy()
    subtr.prune( [x.name for x in node_list] )

    if output_f is None:
        return subtr

    ### Visualize tree
    title=f"Hierarchical clustering of {len(node_list)} factors"

    # Node style
    istyle = NodeStyle()
    istyle["size"] = 0
    for n in subtr.traverse():
        n.set_style(istyle)
    if node_color is not None:
        for n in subtr.traverse():
            if n.is_leaf():
                nstyle = NodeStyle()
                nstyle["fgcolor"] = node_color[n.name]
                nstyle['size'] = 10
                n.set_style(nstyle)
    node_anno = {k: " " + str(k) + " ("+weight_anno[k] + "): " + v  for k,v in enumerate(top_gene) }
    def layout(node):
        if node.is_leaf():
            ete3.faces.add_face_to_node(ete3.TextFace(node_anno[int(node.name)]), node, column=0)

    # Tree style
    ts = TreeStyle()
    ts.layout_fn=layout
    ts.show_leaf_name = False
    ts.show_branch_length = False
    ts.show_scale = False
    if circle:
        ts.mode = "c"
        ts.arc_start = 0
        ts.arc_span = 360
    else:
        if vertical:
            ts.rotation = 90
        ts.branch_vertical_margin = 20

    ts.title.add_face(ete3.TextFace(title, fsize=25),column=0)
    subtr.render(output_f, w=2560, units='mm', tree_style=ts)

    return subtr

def image_to_base64(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")



#! /usr/bin/python

''' helper functions '''
import numpy as np
import pandas as pd
import copy, re, os, geojson
from scipy import sparse
from scipy.special import gammaln, psi, logsumexp, expit, logit
from sklearn.preprocessing import normalize
import scipy.optimize
import sklearn.cluster

from matplotlib.patches import Rectangle
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
from matplotlib.path import Path

def dirichlet_expectation(alpha):
    """
    For a vector theta ~ Dir(alpha), computes E[log(theta)] given alpha.
    """
    assert alpha.min() > 0, "Expecting positive Dirichlet parameters"
    if (len(alpha.shape) == 1):
        return(psi(alpha) - psi(np.sum(alpha)))
    return(psi(alpha) - psi(np.sum(alpha, axis=1)).reshape((-1, 1)))

def pg_mean(b,c=0):
    if np.isscalar(c) and c == 0:
        return b/4
    if not np.isscalar(c):
        if np.isscalar(b):
            b = np.ones(c.shape[0]) * b
        v = b/4
        indx = (c != 0)
        v[indx] = b[indx]/2/c[indx] * np.tanh(c[indx]/2)
        return v
    else:
        return (b/2/c * np.tanh(c/2))

def real_to_sb(mtx):
    assert len(mtx.shape) == 2, "Invalid matrix"
    n, K = mtx.shape
    phi = expit(mtx)
    s = phi[:, 0]
    for k in range(1, K-1):
        phi[:, k] = phi[:, k] * (1. - s)
        s +=  phi[:, k]
    phi[:, K - 1] = 1. - s
    phi = np.clip(phi, 1e-8, 1.-1e-8)
    phi = normalize(phi, norm='l1', axis=1)
    return phi

def sb_to_real(phi):
    assert len(phi.shape) == 2, "Invalid matrix"
    phi = np.clip(phi, 1e-8, 1.-1e-8)
    phi = normalize(phi, norm='l1', axis=1) * (1.-1e-6)
    n, K = phi.shape
    mtx = np.zeros((n, K))
    mtx[:, 0] = logit(phi[:, 0])
    s = phi[:, 0]
    for k in range(1, K):
        mtx[:, k] = logit(phi[:, k] / (1. - s) )
        s += phi[:, k]
    return mtx

def logsumexp_csr(X):
    assert sparse.issparse(X), "logsumexp_csr: invalid input"
    X = X.tocsr()
    result = np.zeros(len(X.data))
    for i in range(X.shape[0]):
        result[X.indptr[i]:X.indptr[i+1]] =\
            logsumexp(X.data[X.indptr[i]:X.indptr[i+1]])
    return result

def gen_even_slices(n, n_packs):
    start = 0
    if n_packs < 1:
        raise ValueError("gen_even_slices got n_packs=%s, must be >=1" % n_packs)
    for pack_num in range(n_packs):
        this_n = n // n_packs
        if pack_num < n % n_packs:
            this_n += 1
        if this_n > 0:
            end = start + this_n
            yield np.arange(start, end)
            start = end

def get_string_with_integer_suff(in_array):
    out = []
    for u in in_array:
        v = re.match('^[A-Za-z]*_*(\d+)$', u.strip())
        if v:
            out.append(v.group(0))
    return out

def get_integer_suff_from_string(in_array):
    out = []
    for u in in_array:
        v = re.match('^[A-Za-z]*_*(\d+)$', u.strip())
        if v:
            out.append(v.group(1))
    return out

def extract_polygons_from_json(gj):
    if isinstance(gj, str):
        assert os.path.exists(gj), f"Invalid file\n{gj}"
        gj = geojson.load(open(gj,'rb'))
    assert gj['type'] == 'FeatureCollection', "Invalid geojson file"
    vertices = []
    for geom in gj['features']:
        if geom['geometry']['type'] == 'Polygon':
            poly = geom['geometry']['coordinates']
            vertices.append(np.array(poly).squeeze())
        elif geom['geometry']['type'] == 'MultiPolygon':
            for poly in geom['geometry']['coordinates']:
                vertices.append(np.array(poly).squeeze())
    return vertices

def svg_parse_list(path):
    commands = {'M': (Path.MOVETO,),'L': (Path.LINETO,),
                'Q': (Path.CURVE3,)*2,'C': (Path.CURVE4,)*3,
                'Z': (Path.CLOSEPOLY,) }
    vertices = []
    codes = []
    for v in path:
        cmd = v[0].upper()
        if len(v) > 1:
            points = [float(x) for x in v[1:]]
            points = np.array(points).reshape((len(points)//2,2))
            vertices.extend(points.tolist())
            codes.extend(commands[cmd])
    return codes, vertices

def match_factors(mtx1, mtx2, c1, n, cmap, mode='beta'):
    """
    Harmonize a pair of LDA results
    """
    k1 = mtx1.shape[0]
    k2 = mtx2.shape[0]
    assert mtx1.shape[1] == mtx2.shape[1], "Invalid input matrices"

    # Try to match two sets of factors
    Q = np.zeros((k1, k2))
    if mode == 'beta':
        beta1 = mtx1 / mtx1.sum(axis = 1).reshape((-1, 1))
        beta2 = mtx2 / mtx2.sum(axis = 1).reshape((-1, 1))
        for k in range(k1):
            q,r = scipy.optimize.nnls(beta2.transpose(), beta1[k, :].transpose())
            Q[k, ] = q
            indx2 = [(-Q[:,k]).argsort() for k in range(k2)]
            indx1 = [(-Q[k,:]).argsort() for k in range(k1)]
    else:
        for k in range(k1):
            q,r = scipy.optimize.nnls(mtx2.transpose(), mtx1[k, :].transpose())
            Q[k, ] = q
            indx2 = [(-Q[:,k]).argsort() for k in range(k2)]
            indx1 = [(-Q[k,:]).argsort() for k in range(k1)]
            Q = Q * mtx2.sum(axis = 1).reshape((1, -1))
            Q = Q / (Q.sum(axis = 1).reshape((-1, 1)))

    model = sklearn.cluster.SpectralBiclustering(n_clusters=(k1,k1), method="bistochastic").fit(Q+1e-5)
    prio = (-Q.sum(axis = 0)).argsort() # sorted from most consistent to less consistent
    step = n//(k1*2)

    dup = [1] * len(c1)
    c1_ref = copy.copy(c1)
    c2 = [-1]*k2
    for i,k in enumerate(prio):
        c2[k] = c1_ref[indx2[k][0]]
        even = dup[indx2[k][0]] % 2
        odd  = dup[indx2[k][0]] + even
        c1_ref[indx2[k][0]] += int(step * (-1)**(even) * (1-1/odd))
        dup[indx2[k][0]] += 1

    indx1 = np.array(c1).argsort()
    v1 = mtx1.sum(axis = 1)
    v1 = v1/v1.sum()
    v1 = v1[indx1]

    indx2 = np.array(c2).argsort()
    v2 = mtx2.sum(axis = 1)
    v2 = v2/v2.sum()
    v2 = v2[indx2]

    st = []
    ed = []
    st_y = []
    ed_y = []
    cl = []

    st_y += [0] + list(np.cumsum(v1)[:-1])
    ed_y += list(np.cumsum(v1))
    cl += [c1[x] for x in indx1]

    offset = 0
    st += [k1+offset]*k1
    ed += [k1+offset]*k1

    st_y += [0] + list(np.cumsum(v2)[:-1])
    ed_y += list(np.cumsum(v2))
    cl += [c2[x] for x in indx2]

    st += [k2-offset]*k2
    ed += [k2-offset]*k2

    df = pd.DataFrame({'xst':st,'yst':st_y,'xed':ed,'yed':ed_y,'Color':cl})

    mid1 = []
    mid2 = []
    v = np.cumsum(v1)
    mid1+=list(v - v1*0.5)
    v = np.cumsum(v2)
    mid2+=list(v - v2*0.5)

    xst = []
    xed = []
    yst = []
    yed = []
    sl = []
    cl = []
    offset = 0.3
    # for ii,i in enumerate(indx1):
    for ii,i in enumerate(indx1):
        for jj,j in enumerate(indx2):
            if Q[i,j] > 1e-3:
                xst.append(k2-offset)
                xed.append(k1+offset)
                yst.append(mid2[jj])
                yed.append(mid1[ii])
                sl.append(Q[i, j])
                cl.append(c1[i])
    df_arrow = pd.DataFrame({'xst':xst,'yst':yst,'xed':xed,'yed':yed,'Weight':sl,'Color':cl})
    return df, df_arrow, c2

def plot_colortable(colors, title, sort_colors=True, ncols=4, dpi = 80,\
                    cell_width = 212, cell_height = 22,\
                    title_fontsize=24, text_fontsize=24,\
                    swatch_width = 48, margin = 12, topmargin = 40):
    # Sort colors by hue, saturation, value and name.
    if sort_colors is True:
        by_hsv = sorted((tuple(mcolors.rgb_to_hsv(mcolors.to_rgb(color))),
                         name)
                        for name, color in colors.items())
        names = [name for hsv, name in by_hsv]
    else:
        names = list(colors)

    n = len(names)
    nrows = n // ncols + int(n % ncols > 0)

    width = cell_width * ncols + 2 * margin
    height = cell_height * nrows + margin + topmargin


    fig, ax = plt.subplots(figsize=(width / dpi, height / dpi), dpi=dpi)
    fig.subplots_adjust(margin/width, margin/height,
                        (width-margin)/width, (height-topmargin)/height)
    ax.set_xlim(0, cell_width * ncols)
    ax.set_ylim(cell_height * (nrows-0.5), -cell_height/2.)
    ax.yaxis.set_visible(False)
    ax.xaxis.set_visible(False)
    ax.set_axis_off()
    ax.set_title(title, fontsize=title_fontsize, loc="left", pad=10)

    for i, name in enumerate(names):
        row = i % nrows
        col = i // nrows
        y = row * cell_height

        swatch_start_x = cell_width * col
        text_pos_x = cell_width * col + swatch_width + 7

        ax.text(text_pos_x, y, name, fontsize=text_fontsize,
                horizontalalignment='left',
                verticalalignment='center')

        ax.add_patch(
            Rectangle(xy=(swatch_start_x, y-9), width=swatch_width,
                      height=18, facecolor=colors[name], edgecolor='0.7')
        )

    return fig



import torch
import pymde
import numpy as np
import matplotlib.colors
import matplotlib.pyplot as plt
import re, os, sys

# A constraint class that implements the unit circle constraint
class circle(pymde.constraints.Constraint):
    def name(self):
        return "circle"

    def initialization(self, n_items, embedding_dim, device=None):
        assert embedding_dim == 2, "Embedding dimension must be 2"
        angles = torch.rand(n_items, device=device) * 2 * torch.tensor([3.141592653589793])
        x_coords = torch.cos(angles)
        y_coords = torch.sin(angles)
        return torch.stack((x_coords, y_coords), dim=1)

    def project_onto_tangent_space(self, X, Z, inplace=True):
        assert Z.shape == X.shape, "Z and X must have the same shape"
        if not torch.all(torch.isclose(torch.norm(X, dim=1), \
                         torch.tensor(1.0), atol=1e-6)):
            X = self.project_onto_constraint(X, inplace=True)
        assert torch.all(torch.isclose(torch.norm(X, dim=1), \
                         torch.tensor(1.0), atol=1e-6)), "X must lie on the unit circle"
        # Compute the dot product of Z and X
        dot_product = torch.sum(Z * X, dim=1, keepdim=True)
        # Subtract the result from Z to get the orthogonal component
        if inplace:
            Z.sub_(dot_product * X)
            return Z
        else:
            return Z - dot_product * X

    def project_onto_constraint(self, Z, inplace=True):
        if inplace:
            Z.div_(torch.norm(Z, dim=1).unsqueeze(-1))
            return Z
        else:
            return torch.div(Z, torch.norm(Z, dim=1).unsqueeze(-1))

# Map pairwise factor proximity to an unit circle embedding then to RGB colors
def assign_color_mds_circle(mtx, cmap_name, weight=None, top_color=None):

    # mtx is a K by K similarity/proximity matrix
    assert mtx.shape[0] == mtx.shape[1], "mtx must be square"
    K = mtx.shape[0]
    # weight is a K vector of factor abundance
    if weight is None:
        weight = np.ones(K)
    weight /= weight.sum()

    # The color of the top factor (the one with the largest weight)
    if top_color is None:
        top_color = "#fcd217"
    else:
        match = re.search(r'^#(?:[0-9a-fA-F]{3}){1,2}$', top_color)
        if match is None:
            top_color = "#fcd217"

    # Construct cost function (want to match large values in mtx to large embedding distances)
    c_dist = []
    c_edge = []
    for i in range(K-1):
        for j in range(i+1,K):
            c_dist.append(-mtx[i,j] * (weight[i]+weight[j]) )
            c_edge.append([i,j])
    f=pymde.penalties.Log(torch.tensor(c_dist, dtype=torch.float32) )

    mde = pymde.MDE(
        n_items=K,
        embedding_dim=2,
        edges=torch.tensor(c_edge, dtype=torch.int64),
        distortion_function=f,
        constraint=circle())
    mde_res = mde.embed().detach().numpy()

    # Recover angles from the embedding
    angle = np.arctan2(mde_res[:, 1], mde_res[:,0]) + np.pi # (0, 2pi)
    anchor_k = np.argmax(weight)
    # Find the offset to map the top factor to the desired color
    cgrid = 200
    cmtx=plt.get_cmap(cmap_name)(np.arange(cgrid)/cgrid)

    top_color_rgb = matplotlib.colors.to_rgb(top_color)
    d = np.abs(cmtx[:, :3] - np.array(top_color_rgb).reshape((1, -1)) ).sum(axis = 1)
    anchor_pos = d.argmin() / cgrid
    anchor_angle = anchor_pos * 2 * np.pi
    angle_shift = angle + (anchor_angle - angle[anchor_k])
    if angle_shift.max() > 2*np.pi:
        angle_shift -= np.pi * 2
    angle_shift[angle_shift < 0] = 2 * np.pi + angle_shift[angle_shift < 0]
    c_pos = angle_shift / np.pi / 2
    return c_pos



import sys, os, copy, gzip, logging
import pickle, argparse
import numpy as np
import pandas as pd

from scipy.sparse import *
import sklearn.neighbors
import sklearn.preprocessing
from sklearn.decomposition import LatentDirichletAllocation as LDA

def lda(_args):

    parser = argparse.ArgumentParser(prog = "lda")

    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--unit_label', default = 'random_index', type=str, help='Which column to use as unit identifier')
    parser.add_argument('--unit_attr', type=str, nargs='+', default=[], help='')
    parser.add_argument('--feature', type=str, default='', help='')
    parser.add_argument('--feature_label', default = "gene", type=str, help='Which column to use as feature identifier')
    parser.add_argument('--key', default = 'count', type=str, help='')
    parser.add_argument('--train_on', default = '', type=str, help='')

    parser.add_argument('--nFactor', type=int, default=10, help='')
    parser.add_argument('--minibatch_size', type=int, default=512, help='')
    parser.add_argument('--min_ct_per_feature', type=int, default=1, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=20, help='')
    parser.add_argument('--thread', type=int, default=1, help='')
    parser.add_argument('--epoch', type=int, default=1, help='How many times to loop through the full data')
    parser.add_argument('--epoch_id_length', type=int, default=-1, help='')
    parser.add_argument('--use_model', type=str, default='', help="Use provided model to transform input data")
    parser.add_argument('--overwrite', action='store_true')

    args = parser.parse_args(_args)
    logger = logging.getLogger(__name__)

    if args.use_model != '' and not os.path.exists(args.use_model):
        sys.exit("Invalid model file")

    unit_attr = [x.lower() for x in args.unit_attr]
    key = args.key.lower()
    unit_key = args.unit_label.lower()
    gene_key = args.feature_label.lower()
    train_on = args.train_on.lower()
    if train_on == '':
        train_on = key
    adt = {unit_key:str, gene_key:str, key:int}
    adt.update({x:str for x in unit_attr})

    ### Basic parameterse
    b_size = args.minibatch_size
    K = args.nFactor

    ### Input
    # Required columns: unit ID, gene, key
    required_header = [unit_key,gene_key,train_on]
    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file.")
    with gzip.open(args.input, 'rt') as rf:
        header = rf.readline().strip().split('\t')
    header = [x.lower() for x in header]
    for x in required_header:
        if x not in header:
            sys.exit("Input file must have at least 3 columns: unit label, feature label, count, matching the customized column names (case insensitive) --unit_label, --feature_label, and --key/--train_on")

    use_existing_model = False
    model_f = args.output+".model.p"
    if os.path.exists(args.use_model):
        model_f = args.use_model
    if not args.overwrite and os.path.exists(model_f):
        lda = pickle.load( open( model_f, "rb" ) )
        feature_kept = lda.feature_names_in_
        lda.feature_names_in_ = None
        ft_dict = {x:i for i,x in enumerate( feature_kept ) }
        K, M = lda.components_.shape
        use_existing_model = True
        logger.warning(f"Read existing model from\n{model_f}\n use --overwrite to allow the model files to be overwritten\n{M} genes will be used")

    factor_header = [str(x) for x in range(K)]
    if not use_existing_model:
        if not os.path.exists(args.feature):
            sys.exit("Unable to read feature list")
        ### Use only the provided list of features
        with gzip.open(args.feature, 'rt') as rf:
            fheader = rf.readline().strip().split('\t')
        fheader = [x.lower() for x in fheader]
        feature=pd.read_csv(args.feature, sep='\t', skiprows=1, names=fheader, dtype={gene_key:str, key:int})
        feature = feature[feature[key] >= args.min_ct_per_feature]
        feature.sort_values(by=key,ascending=False,inplace=True)
        feature.drop_duplicates(subset=gene_key,keep='first',inplace=True)
        feature_kept = list(feature[gene_key].values)
        ft_dict = {x:i for i,x in enumerate( feature_kept ) }
        M = len(feature_kept)

        logger.info(f"Start fitting model ... model will be stored in\n{model_f}\n{M} genes will be used")
        lda = LDA(n_components=K, learning_method='online', batch_size=b_size, n_jobs = args.thread, verbose = 0)
        feature_mf = np.array(feature[key].values).astype(float)
        feature_mf/= feature_mf.sum()
        epoch = 0
        n_unit = 0
        epoch_id = set()
        while epoch < args.epoch:
            df = pd.DataFrame()
            for chunk in pd.read_csv(gzip.open(args.input, 'rt'), \
                    sep='\t',chunksize=1000000, skiprows=1, names=header, \
                    usecols=[unit_key,gene_key,train_on], dtype=adt):
                chunk = chunk[chunk[gene_key].isin(feature_kept)]
                if args.epoch_id_length > 0:
                    v = set(chunk[unit_key].map(lambda x: x[:args.epoch_id_length]).unique())
                    epoch_id.update(v)
                    epoch = len(epoch_id) - 1
                chunk.rename(columns = {train_on:key}, inplace=True)
                if chunk.shape[0] == 0:
                    continue
                last_indx = chunk[unit_key].iloc[-1]
                df = pd.concat([df, chunk[~chunk[unit_key].eq(last_indx)]])
                if len(df[unit_key].unique()) < b_size * 1.5: # Left to next chunk
                    df = pd.concat((df, chunk[chunk[unit_key].eq(last_indx)]))
                    continue
                # Total mulecule count per unit
                brc = df.groupby(by = [unit_key]).agg({key: sum}).reset_index()
                brc = brc[brc[key] > args.min_ct_per_unit]
                brc.index = range(brc.shape[0])
                df = df[df[unit_key].isin(brc[unit_key].values)]
                # Make DGE
                barcode_kept = list(brc[unit_key].values)
                bc_dict  = {x:i for i,x in enumerate( barcode_kept ) }
                indx_row = [ bc_dict[x] for x in df[unit_key]]
                indx_col = [ ft_dict[x] for x in df[gene_key]]
                N = len(barcode_kept)
                mtx = coo_matrix((df[key].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
                x1 = np.median(brc[key].values)
                x2 = np.mean(brc[key].values)
                logger.info(f"Made DGE {mtx.shape}, median/mean count: {x1:.1f}/{x2:.1f}")
                _ = lda.partial_fit(mtx)

                df = copy.copy(chunk[chunk[unit_key].eq(last_indx)] )
                logl = lda.score(mtx) / mtx.shape[0]
                n_unit += N
                logger.info(f"Epoch {epoch}, finished {n_unit} units. batch logl: {logl:.4f}")
                if epoch >= args.epoch:
                    break

            # Leftover
            if len(df[unit_key].unique()) > b_size:
                brc = df.groupby(by = [unit_key]).agg({key: sum}).reset_index()
                brc = brc[brc[key] > args.min_ct_per_unit]
                brc.index = range(brc.shape[0])
                df = df[df[unit_key].isin(brc[unit_key].values)]
                barcode_kept = list(brc[unit_key].values)
                bc_dict  = {x:i for i,x in enumerate( barcode_kept ) }
                indx_row = [ bc_dict[x] for x in df[unit_key]]
                indx_col = [ ft_dict[x] for x in df[gene_key]]
                N = len(barcode_kept)
                mtx = coo_matrix((df[key].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
                x1 = np.median(brc[key].values)
                x2 = np.mean(brc[key].values)
                logger.info(f"Made DGE {mtx.shape}, median/mean count: {x1:.1f}/{x2:.1f}")
                _ = lda.partial_fit(mtx)
                logl = lda.score(mtx) / mtx.shape[0]
                logger.info(f"logl: {logl:.4f}")

            if args.epoch_id_length <= 0:
                epoch += 1

        lda.feature_names_in_ = feature_kept
        # Relabel factors based on (approximate) descending abundance
        weight = lda.components_.sum(axis=1)
        ordered_k = np.argsort(weight)[::-1]
        lda.components_ = lda.components_[ordered_k,:]
        lda.exp_dirichlet_component_ = lda.exp_dirichlet_component_[ordered_k,:]
        # Store model
        pickle.dump( lda, open( model_f, "wb" ) )
        out_f = model_f.replace("model.p", "model_matrix.tsv.gz")
        pd.concat([pd.DataFrame({gene_key: lda.feature_names_in_}),\
                    pd.DataFrame(lda.components_.T,\
                    columns = [str(k) for k in range(K)], dtype='float64')],\
                    axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.4e', compression={"method":"gzip"})


    ###
    ### Rerun all units once and store results
    ###
    dtp = {'topK':int,key:int,unit_key:str}
    dtp.update({x:float for x in ['topP']+factor_header})
    res_f = args.output+".fit_result.tsv.gz"
    nbatch = 0
    logger.info(f"Result file {res_f}")

    if train_on != key:
        post_count = np.zeros((K, M))
        epoch_id = ''
        end_of_epoch = False
        df = pd.DataFrame()
        for chunk in pd.read_csv(gzip.open(args.input, 'rt'), \
                sep='\t',chunksize=1000000, skiprows=1, names=header, \
                usecols=[unit_key,gene_key,key,train_on]+unit_attr, dtype=adt):
            chunk = chunk[chunk[gene_key].isin(feature_kept)]
            if chunk.shape[0] == 0:
                continue
            if args.epoch_id_length > 0:
                v = chunk[unit_key].map(lambda x: x[:args.epoch_id_length]).values
                if epoch_id == '':
                    epoch_id = v[0]
                if v[-1] != v[0]:
                    end_of_epoch = True
                    chunk = chunk.loc[v == epoch_id, :]
                if chunk.shape[0] == 0:
                    continue
            last_indx = chunk[unit_key].iloc[-1]
            df = pd.concat([df, chunk[~chunk[unit_key].eq(last_indx)]])
            if len(df[unit_key].unique()) < b_size * 1.5: # Left to next chunk
                df = pd.concat((df, chunk[chunk[unit_key].eq(last_indx)]))
                continue
            # Total mulecule count per unit
            brc = df.groupby(by = [unit_key]).agg({key: sum}).reset_index()
            brc = brc.merge(right=df[[unit_key]+unit_attr].drop_duplicates(subset=unit_key),on=unit_key,how='inner')
            brc = brc[brc[key] > args.min_ct_per_unit]
            brc.index = range(brc.shape[0])
            df = df[df[unit_key].isin(brc[unit_key].values)]

            # Make training DGE
            barcode_kept = list(brc[unit_key].values)
            bc_dict  = {x:i for i,x in enumerate( barcode_kept ) }
            indx_row = [ bc_dict[x] for x in df[unit_key]]
            indx_col = [ ft_dict[x] for x in df[gene_key]]
            N = len(barcode_kept)
            mtx = coo_matrix((df[train_on].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
            x1 = np.median(brc[key].values)
            x2 = np.mean(brc[key].values)
            logger.info(f"Made DGE {mtx.shape}, median/mean count: {x1:.1f}/{x2:.1f}")

            theta = lda.transform(mtx)

            # Aggregate testing counts
            test_ct = df[key].values - df[train_on].values
            if np.min(test_ct) < 0:
                sys.exit("Total count cannot be smaller than training count")
            mtx = coo_matrix((test_ct, (indx_row, indx_col)), shape=(N, M)).tocsr()
            post_count += np.array(theta.T @ mtx)

            brc = pd.concat((brc, pd.DataFrame(theta, columns = factor_header)), axis = 1)
            brc['topK'] = np.argmax(theta, axis = 1).astype(int)
            brc['topP'] = np.max(theta, axis = 1)
            brc = brc.astype(dtp)
            logger.info(f"{nbatch}-th batch with {brc.shape[0]} units")
            if nbatch == 0:
                brc.to_csv(res_f, sep='\t', mode='w', float_format="%.4e", index=False, header=True, compression={"method":"gzip"})
            else:
                brc.to_csv(res_f, sep='\t', mode='a', float_format="%.4e", index=False, header=False, compression={"method":"gzip"})
            nbatch += 1
            df = copy.copy(chunk[chunk[unit_key].eq(last_indx)] )
            if end_of_epoch:
                break

        # Leftover
        brc = df.groupby(by = [unit_key]).agg({key: sum}).reset_index()
        brc = brc.merge(right=df[[unit_key]+unit_attr].drop_duplicates(subset=unit_key),on=unit_key,how='inner')
        brc = brc[brc[key] > args.min_ct_per_unit]
        brc.index = range(brc.shape[0])
        # print(brc.shape)
        if brc.shape[0] > 0:
            df = df[df[unit_key].isin(brc[unit_key].values)]
            # Make training DGE
            barcode_kept = list(brc[unit_key].values)
            bc_dict  = {x:i for i,x in enumerate( barcode_kept ) }
            indx_row = [ bc_dict[x] for x in df[unit_key]]
            indx_col = [ ft_dict[x] for x in df[gene_key]]
            N = len(barcode_kept)
            mtx = coo_matrix((df[train_on].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
            x1 = np.median(brc[key].values)
            x2 = np.mean(brc[key].values)
            logger.info(f"Made DGE {mtx.shape}, median/mean count: {x1:.1f}/{x2:.1f}")

            theta = lda.transform(mtx)

            # Aggregate testing counts
            test_ct = df[key].values - df[train_on].values
            if np.min(test_ct) < 0:
                sys.exit("Total count cannot be smaller than training count")
            mtx = coo_matrix((test_ct, (indx_row, indx_col)), shape=(N, M)).tocsr()
            post_count += np.array(theta.T @ mtx)

            brc = pd.concat((brc, pd.DataFrame(theta, columns = factor_header)), axis = 1)
            brc['topK'] = np.argmax(theta, axis = 1).astype(int)
            brc['topP'] = np.max(theta, axis = 1)
            brc = brc.astype(dtp)
            logger.info(f"{nbatch}-th batch with {brc.shape[0]} units")
            if nbatch == 0:
                brc.to_csv(res_f, sep='\t', mode='w', float_format="%.4e", index=False, header=True, compression={"method":"gzip"})
            else:
                brc.to_csv(res_f, sep='\t', mode='a', float_format="%.4e", index=False, header=False, compression={"method":"gzip"})

        logger.info(f"Finished ({nbatch})")

        out_f = args.output+".posterior.count.tsv.gz"
        pd.concat([pd.DataFrame({gene_key: feature_kept}),\
                pd.DataFrame(post_count.T, dtype='float64',\
                                columns = [str(k) for k in range(K)])],\
                axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})

        sys.exit()





    post_count = np.zeros((K, M))
    epoch_id = ''
    df = pd.DataFrame()
    end_of_epoch = False
    for chunk in pd.read_csv(gzip.open(args.input, 'rt'), \
            sep='\t',chunksize=1000000, skiprows=1, names=header, \
            usecols=[unit_key,gene_key,key]+unit_attr, dtype=adt):
        chunk = chunk[chunk[gene_key].isin(feature_kept)]
        if chunk.shape[0] == 0:
            continue
        if args.epoch_id_length > 0:
            v = chunk[unit_key].map(lambda x: x[:args.epoch_id_length]).values
            if epoch_id == '':
                epoch_id = v[0]
            if v[-1] != v[0]:
                end_of_epoch = True
                chunk = chunk.loc[v == epoch_id, :]
            if chunk.shape[0] == 0:
                continue
        last_indx = chunk[unit_key].iloc[-1]
        df = pd.concat([df, chunk[~chunk[unit_key].eq(last_indx)]])
        if len(df[unit_key].unique()) < b_size * 1.5: # Left to next chunk
            df = pd.concat((df, chunk[chunk[unit_key].eq(last_indx)]))
            continue
        # Total mulecule count per unit
        brc = df.groupby(by = [unit_key]).agg({key: sum}).reset_index()
        brc = brc.merge(right=df[[unit_key]+unit_attr].drop_duplicates(subset=unit_key),on=unit_key,how='inner')
        brc = brc[brc[key] > args.min_ct_per_unit]
        brc.index = range(brc.shape[0])
        df = df[df[unit_key].isin(brc[unit_key].values)]

        # Make training DGE
        barcode_kept = list(brc[unit_key].values)
        bc_dict  = {x:i for i,x in enumerate( barcode_kept ) }
        indx_row = [ bc_dict[x] for x in df[unit_key]]
        indx_col = [ ft_dict[x] for x in df[gene_key]]
        N = len(barcode_kept)
        mtx = coo_matrix((df[key].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
        x1 = np.median(brc[key].values)
        x2 = np.mean(brc[key].values)
        logger.info(f"Made DGE {mtx.shape}, median/mean count: {x1:.1f}/{x2:.1f}")

        theta = lda.transform(mtx)
        post_count += np.array(theta.T @ mtx)

        brc = pd.concat((brc, pd.DataFrame(theta, columns = factor_header)), axis = 1)
        brc['topK'] = np.argmax(theta, axis = 1).astype(int)
        brc['topP'] = np.max(theta, axis = 1)
        brc = brc.astype(dtp)
        logger.info(f"{nbatch}-th batch with {brc.shape[0]} units")
        if nbatch == 0:
            brc.to_csv(res_f, sep='\t', mode='w', float_format="%.4e", index=False, header=True, compression={"method":"gzip"})
        else:
            brc.to_csv(res_f, sep='\t', mode='a', float_format="%.4e", index=False, header=False, compression={"method":"gzip"})
        nbatch += 1
        df = copy.copy(chunk[chunk[unit_key].eq(last_indx)] )
        if end_of_epoch:
            break

    # Leftover
    brc = df.groupby(by = [unit_key]).agg({key: sum}).reset_index()
    brc = brc.merge(right=df[[unit_key]+unit_attr].drop_duplicates(subset=unit_key),on=unit_key,how='inner')
    brc = brc[brc[key] > args.min_ct_per_unit]
    brc.index = range(brc.shape[0])
    if brc.shape[0] > 0:
        df = df[df[unit_key].isin(brc[unit_key].values)]
        # Make DGE
        barcode_kept = list(brc[unit_key].values)
        bc_dict  = {x:i for i,x in enumerate( barcode_kept ) }
        indx_row = [ bc_dict[x] for x in df[unit_key]]
        indx_col = [ ft_dict[x] for x in df[gene_key]]
        N = len(barcode_kept)
        mtx = coo_matrix((df[key].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
        x1 = np.median(brc[key].values)
        x2 = np.mean(brc[key].values)
        logger.info(f"Made DGE {mtx.shape}, median/mean count: {x1:.1f}/{x2:.1f}")

        theta = lda.transform(mtx)
        post_count += np.array(theta.T @ mtx)

        brc = pd.concat((brc, pd.DataFrame(theta, columns = factor_header)), axis = 1)
        brc['topK'] = np.argmax(theta, axis = 1).astype(int)
        brc['topP'] = np.max(theta, axis = 1)
        brc = brc.astype(dtp)
        if nbatch == 0:
            brc.to_csv(res_f, sep='\t', mode='w', float_format="%.4e", index=False, header=True, compression={"method":"gzip"})
        else:
            brc.to_csv(res_f, sep='\t', mode='a', float_format="%.4e", index=False, header=False, compression={"method":"gzip"})

    logger.info(f"Finished ({nbatch})")

    out_f = args.output+".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({gene_key: feature_kept}),\
            pd.DataFrame(post_count.T, dtype='float64',\
                            columns = [str(k) for k in range(K)])],\
            axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})



### Simple differential expression tests

import sys, io, os, gzip, copy, re, time, pickle, argparse
import numpy as np
import pandas as pd
import scipy.stats
from scipy.sparse import *
from joblib.parallel import Parallel, delayed

from ficture.utils import utilt

def de_bulk(_args):

    parser = argparse.ArgumentParser(prog="de_bulk")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--feature', type=str, default='', help='')
    parser.add_argument('--feature_label', type=str, default = "gene", help='')
    parser.add_argument('--min_ct_per_feature', default=50, type=int, help='')
    parser.add_argument('--max_pval_output', default=1e-3, type=float, help='')
    parser.add_argument('--min_fold_output', default=1.5, type=float, help='')
    parser.add_argument('--min_output_per_factor', default=10, type=int, help='Even when there are no significant DE genes, output top genes for each factor')
    parser.add_argument('--thread', default=1, type=int, help='')
    args = parser.parse_args(_args)

    pcut=args.max_pval_output
    fcut=args.min_fold_output
    gene_kept = set()
    if os.path.exists(args.feature):
        feature = pd.read_csv(args.feature, sep='\t', header=0)
        gene_kept = set(feature[args.feature_label].values )

    # Read aggregated count table
    info = pd.read_csv(args.input,sep='\t',header=0)
    oheader = []
    header = []
    for x in info.columns:
        y = re.match('^[A-Za-z]*_*(\d+)$', x)
        if y:
            header.append(y.group(1))
            oheader.append(x)
    K = len(header)
    M = info.shape[0]
    reheader = {oheader[k]:header[k] for k in range(K)}
    reheader[args.feature_label] = "gene"
    info.rename(columns = reheader, inplace=True)
    print(f"Read posterior count over {M} genes and {K} factors")

    if len(gene_kept) > 0:
        info = info.loc[info.gene.isin(gene_kept), :]
    info["gene_tot"] = info.loc[:, header].sum(axis=1).astype(int)
    info = info[info["gene_tot"] > args.min_ct_per_feature]
    info.index = info.gene.values
    total_umi = info.gene_tot.sum()
    total_k = np.array(info.loc[:, [str(k) for k in range(K)]].sum(axis = 0) )
    M = info.shape[0]

    print(f"Testing {M} genes over {K} factors")

    def chisq(k,info,total_k,total_umi):
        res = []
        if total_k <= 0:
            return res
        for name, v in info.iterrows():
            if v[str(k)] <= 0:
                continue
            tab=np.zeros((2,2))
            tab[0,0]=v[str(k)]
            tab[0,1]=v["gene_tot"]-tab[0,0]
            tab[1,0]=total_k-tab[0,0]
            tab[1,1]=total_umi-total_k-v["gene_tot"]+tab[0,0]
            fd=tab[0,0]/total_k/tab[0,1]*(total_umi-total_k)
            if fd < 1:
                continue
            tab = np.around(tab, 0).astype(int) + 1
            chi2, p, dof, ex = scipy.stats.chi2_contingency(tab, correction=False)
            res.append([name,k,chi2,p,fd,v["gene_tot"]])
        return res

    res = []
    if args.thread > 1:
        for k in range(K):
            idx_slices = [idx for idx in utilt.gen_even_slices(M, args.thread)]
            with Parallel(n_jobs=args.thread, verbose=0) as parallel:
                result = parallel(delayed(chisq)(k, \
                            info.iloc[idx, :].loc[:, [str(k), 'gene_tot']],\
                            total_k[k], total_umi) for idx in idx_slices)
            res += [item for sublist in result for item in sublist]
    else:
        for name, v in info.iterrows():
            for k in range(K):
                if total_k[k] <= 0 or v[str(k)] <= 0:
                    continue
                tab=np.zeros((2,2))
                tab[0,0]=v[str(k)]
                tab[0,1]=v["gene_tot"]-tab[0,0]
                tab[1,0]=total_k[k]-tab[0,0]
                tab[1,1]=total_umi-total_k[k]-v["gene_tot"]+tab[0,0]
                fd=tab[0,0]/total_k[k]/tab[0,1]*(total_umi-total_k[k])
                if fd < 1:
                    continue
                tab = np.around(tab, 0).astype(int) + 1
                chi2, p, dof, ex = scipy.stats.chi2_contingency(tab, correction=False)
                res.append([name,k,chi2,p,fd,v["gene_tot"]])

    chidf=pd.DataFrame(res,columns=['gene','factor','Chi2','pval','FoldChange','gene_total'])
    chidf.sort_values(by=['factor','Chi2'],ascending=[True,False],inplace=True)
    chidf['Rank'] = 0
    for k in range(K):
        chidf.loc[chidf.factor.eq(k), 'Rank'] = np.arange((chidf.factor==k).sum())
    chidf = chidf.loc[((chidf.pval<pcut)&(chidf.FoldChange>fcut)) | (chidf.Rank < args.min_output_per_factor), :]
    chidf.sort_values(by=['factor','Chi2'],ascending=[True,False])
    chidf.Chi2 = chidf.Chi2.map(lambda x : "{:.1f}".format(x) )
    chidf.FoldChange = chidf.FoldChange.map(lambda x : "{:.2f}".format(x) )
    chidf.gene_total = chidf.gene_total.astype(int)
    chidf.drop(columns = 'Rank', inplace=True)

    outpath=os.path.dirname(args.output)
    if not os.path.exists(outpath):
        os.makedirs(outpath)
    chidf.to_csv(args.output,sep='\t',float_format="%.2e",index=False)



### Read (randomized) hexagons from file, construct minibatch
import sys, os, gzip, copy, re
import numpy as np
import pandas as pd
from scipy.sparse import coo_array

from ficture.models.lda_minibatch import PairedMinibatch, Minibatch

class UnitLoaderAugmented:

    def __init__(self, reader, ft_dict, key, bkey, batch_id_prefix=0, min_ct_per_unit=1, unit_attr=['x','y']) -> None:
        self.reader = reader
        self.ft_dict = ft_dict
        self.key = key
        self.bkey= bkey
        self.df = pd.DataFrame()
        self.file_is_open = True
        self.batch = None
        self.brc = None
        self.prefix = batch_id_prefix
        self.min_ct_per_unit = min_ct_per_unit
        self.unit_attr = list(unit_attr)
        self.M = max(self.ft_dict.values()) + 1
        self.batch_id_list = set()

    def update_batch(self, bsize):
        n_unit = 0
        if len(self.df) > 0:
            n_unit = len(self.df.unit.unique())
        if not self.file_is_open:
            return 0
        while n_unit <= bsize:
            try:
                chunk = next(self.reader)
            except StopIteration:
                self.file_is_open = False
                break
            self.df = pd.concat([self.df, chunk])
            n_unit = len(self.df.unit.unique())
        left = pd.DataFrame()
        if self.file_is_open:
            last_indx = self.df.unit.iloc[-1]
            left = self.df[self.df.unit == last_indx]
            self.df = self.df[self.df.unit != last_indx]
        self.brc = self.df[['unit']+self.unit_attr].drop_duplicates(subset=['unit'])
        self.brc = self.brc.merge(right = self.df.groupby(by='unit').agg({self.key:sum, self.bkey:sum}).reset_index(), on = 'unit', how = 'inner' )
        self.brc = self.brc[self.brc[self.key] >= self.min_ct_per_unit]
        buffer_weight = self.brc[self.bkey].values / (self.brc[self.bkey].values + self.brc[self.key].values)
        barcode_kept=list(self.brc.unit)
        bt_dict={x:i for i,x in enumerate(barcode_kept)}
        N = len(bt_dict)
        self.df = self.df[self.df.unit.isin(bt_dict) & self.df.gene.isin(self.ft_dict)]
        if self.prefix > 0:
            self.batch_id_list.update(set(self.df.unit.map(lambda x : x[:self.prefix]).values))
        self.batch = PairedMinibatch(\
            mtx_focal = coo_array((self.df[self.key].values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr(),
            mtx_buffer = coo_array((self.df[self.bkey].values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr(),
            buffer_weight = buffer_weight)
        self.df = copy.copy(left)
        return self.batch.n



class UnitLoader:

    def __init__(self, reader, ft_dict, key, batch_id_prefix=0, min_ct_per_unit=1, unit_id='unit', unit_attr=['x','y'], debug=False) -> None:
        self.reader = reader
        self.ft_dict = ft_dict
        self.key = key
        self.df = pd.DataFrame()
        self.file_is_open = True
        self.mtx = None
        self.brc = None
        self.prefix = batch_id_prefix
        self.min_ct_per_unit = min_ct_per_unit
        self.unit_id = unit_id
        self.unit_attr = list(unit_attr)
        self.M = max(self.ft_dict.values()) + 1
        self.debug = debug
        self.batch_id_list = set()

    def update_batch(self, bsize):
        n_unit = 0
        if len(self.df) > 0:
            n_unit = len(self.df.unit.unique())
        if not self.file_is_open:
            return 0
        while n_unit <= bsize:
            try:
                chunk = next(self.reader)
                chunk.rename(columns={self.unit_id:'unit'}, inplace=True)
                chunk = chunk[chunk.gene.isin(self.ft_dict)]
                if self.debug:
                    print(f"Read {chunk.shape[0]} lines from file")
            except StopIteration:
                self.file_is_open = False
                break
            self.df = pd.concat([self.df, chunk])
            n_unit = len(self.df.unit.unique())
        left = pd.DataFrame()
        if self.file_is_open:
            last_indx = self.df.unit.iloc[-1]
            left = self.df[self.df.unit == last_indx]
            self.df = self.df[self.df.unit != last_indx]
        self.brc = self.df[['unit']+self.unit_attr].drop_duplicates(subset=['unit'])
        self.brc = self.brc.merge(right = self.df.groupby(by='unit').agg({self.key:sum}).reset_index(), on = 'unit', how = 'inner' )
        self.brc = self.brc[self.brc[self.key] >= self.min_ct_per_unit]
        barcode_kept=list(self.brc.unit)
        bt_dict={x:i for i,x in enumerate(barcode_kept)}
        N = len(bt_dict)
        self.df = self.df[self.df.unit.isin(bt_dict)]
        if self.prefix > 0:
            self.batch_id_list.update(set(self.df.unit.map(lambda x : x[:self.prefix]).values))
        self.mtx = coo_array((self.df[self.key].values, \
                            (self.df.unit.map(bt_dict).values, \
                             self.df.gene.map(self.ft_dict).values) ), \
                            shape=(N, self.M)).tocsr()
        self.df = copy.copy(left)
        return N



import pandas as pd

def layout_map(meta_data, layout, lane = None):

    ### Range / offset
    # Lane & tile ids are treated as string
    # Assume manifest file has an id column with format lane_tile
    mani=pd.read_csv(meta_data, sep='\t')
    mani["lane"] = mani["id"].map(lambda x : x.split('_')[0])
    mani["tile"] = mani["id"].map(lambda x : x.split('_')[1])
    mani["ybin"] = mani.ymax - mani.ymin
    mani["xbin"] = mani.xmax - mani.xmin

    ### Layout
    layout = pd.read_csv(layout, sep='\t', dtype=str)
    for key in ['row', 'col']:
        layout[key] = layout[key].astype(int)

    if lane is not None:
        try:
            lane = [str(lane)]
        except:
            lane = [str(x) for x in list(lane)]
        mani = mani[mani.lane.isin(lane)]
        layout = layout[layout.lane.isin(lane)]

    df = layout.merge(right = mani, on = ["lane", "tile"], how = "left")
    df.sort_values(by = ['lane', 'row', 'col'], inplace=True)
    df.row = df.row - df.row.min()
    df.col = df.col - df.col.min()
    nrows = df.row.max() + 1
    ncols = df.col.max() + 1
    lanes = []
    tiles = []
    for i in range(nrows):
        lanes.append( [None] * ncols )
        tiles.append( [None] * ncols )
    for it, row in df.iterrows():
        i = row['row']
        j = row['col']
        lanes[i][j] = row['lane']
        tiles[i][j] = row['tile']
    df.set_index(["lane", "tile"], inplace=True)

    return df, lanes, tiles



# Functions for fitting Latent Dirichlet Allocation
# with online variational Bayes (VB)
# following Hoffman et al. 2010

import sys, io, os, re, time, copy, warnings
import numpy as np
from scipy.special import gammaln, psi, logsumexp, expit, logit
from scipy.sparse import issparse
from sklearn.preprocessing import normalize
from joblib.parallel import Parallel, delayed

# Add directory
from ficture.utils import utilt

class OnlineLDA:
    """
    Implements online VB for LDA as described in Hoffman et al. 2010.
    """
    def __init__(self, vocab, K, N, alpha = None, eta = None, tau0=9, kappa=.7, iter_inner = 50, tol = 1e-4, verbose = 0, thread = 1):
        """
        Arguments:
        vocab: A list of features
        K:     Number of topics
        N:     Estimated total number of units
        alpha: Hyperparameter for the prior on theta
        eta:   Hyperparameter for prior on beta
        tau0:  A (non-negative) learning parameter that downweights earl iterations
        kappa: Learning rate (exponential decay), should be between
               (0.5, 1.0] to guarantee asymptotic convergence.
        """
        self._vocab = vocab
        self._K = K
        self._M = len(self._vocab)
        self._N = N
        self._tau0 = tau0 + 1
        self._kappa = kappa
        self._updatect = 0
        self._max_iter_inner = iter_inner
        self._tol = tol
        self._eps = 1e-16
        self._verbose = verbose
        self._thread = thread
        self._Elog_beta = None      # K x M
        self._lambda = None         # K x M
        self._expElog_beta = None   # K x M
        self._sstats = None         # K x M

        self._alpha = alpha # Factor weight prior
        if self._alpha is None:
            self._alpha = np.ones(self._K)/self._K
        elif np.isscalar(self._alpha):
            self._alpha = np.ones(self._K)*self._alpha
        elif len(self._alpha.shape) == 1:
            assert self._alpha.shape[0] == self._K, "Invalid alpha"
        else:
            warnings.warn("Invalid alpha, fall back to default uniform 1/K")
            self._alpha = np.ones(self._K)/self._K

        self._eta = eta  # Expression profile prior
        if self._eta is None:
            self._eta = (np.ones(self._M)/self._K).reshape((1,-1))
        elif np.isscalar(self._eta):
            self._eta = (np.ones(self._M)*self._eta).reshape((1,-1))
        elif len(self._eta.shape) == 1:
            self._eta = np.array(self._eta).reshape((1, -1))
            assert self._eta.shape[1] == self._M, "Invalid eta"
        elif len(self._eta.shape) == 2:
            assert self._eta.shape == (self._K, self._M), "Invalid eta"
        else:
            warnings.warn("Invalid eta, fall back to default uniform 1/K")
            self._eta = (np.ones(self._M)/self._K).reshape((1,-1))

    def init_global_parameter(self, _lambda=None):
        # Initialize the variational distribution q(beta|lambda)
        if _lambda is None:
            self._lambda = np.random.gamma(100., 1./100., (self._K, self._M))
        else:
            self._lambda = _lambda
            assert self._lambda.shape == (self._K, self._M), "Invalid lambda"
        if self._lambda.min() <= 0 :
            warnings.warn("Parameters must be positive, will clip to 1/10 the min nonzero value")
            pseudo = self._lambda[self._lambda > 0].min()/10
            self._lambda = np.clip(self._lambda, pseudo, None)
        self._Elog_beta = utilt.dirichlet_expectation(self._lambda)
        self._expElog_beta = np.exp(self._Elog_beta)


    def _update_gamma(self, X, _gamma, alpha):
        gamma = copy.copy(_gamma)
        sstats = np.zeros((self._K, self._M))
        expElog_theta = np.exp(utilt.dirichlet_expectation(gamma))
        for j in range(X.shape[0]):
            maxchange = self._tol + 1
            it = 0
            while it < self._max_iter_inner and maxchange > self._tol:
                old_gamma = copy.copy(gamma[j, :])
                phi_norm = expElog_theta[j, :] @ self._expElog_beta + self._eps # 1 x M
                gamma[j, :] = alpha[j, :] +\
                    np.multiply(expElog_theta[j, :], \
                                (X[[j], :].multiply(1./phi_norm)) @ self._expElog_beta.T) # 1 x K
                expElog_theta[j, :] = np.exp(utilt.dirichlet_expectation(gamma[j, :]))
                maxchange = np.abs(old_gamma/old_gamma.sum() - gamma[j, :]/gamma[j, :].sum()).max()
                it += 1
                if self._verbose > 2 or (self._verbose > 1 and it % 10 == 0):
                    print(f"E-step, {j}-th unit, update gamma: {it}-th iteration, max change {maxchange:.4f}")
            phi_norm = expElog_theta[j, :] @ self._expElog_beta + self._eps
            phi = expElog_theta[j, :].reshape((-1, 1)) * self._expElog_beta / phi_norm.reshape((1, -1)) # K x M
            sstats += np.multiply(X[[j], :].toarray(), phi)
        return sstats, gamma


    def do_e_step(self, batch):
        """
        Update local parameters, compute sufficient stats for M step
        """
        # Initialize the variational distribution q(theta|gamma)
        if batch.alpha is None:
            batch.alpha = np.broadcast_to(self._alpha, (batch.n, self._K))
        if batch.gamma is None:
            batch.gamma = np.random.gamma(100., 1./100., (batch.n, self._K))

        # Run E-step in parallel
        if self._thread <= 1:
            self._sstats, batch.gamma = self._update_gamma(batch.mtx, batch.gamma, batch.alpha)
        else:
            idx_slices = [idx for idx in utilt.gen_even_slices(batch.n, self._thread)]
            with Parallel(n_jobs=self._thread, verbose=max(0, self._verbose - 1)) as parallel:
                result = parallel(delayed(self._update_gamma)(batch.mtx[idx, :], batch.gamma[idx, :], batch.alpha[idx, :])
                for idx in idx_slices)
            # Collect sufficient statistics
            self._sstats = np.zeros((self._K, self._M))
            for i, v in enumerate(result):
                self._sstats += v[0]
                batch.gamma[idx_slices[i], :] = v[1]

        batch.ll = np.multiply(normalize(batch.gamma, norm='l1', axis=1), batch.mtx @ self._Elog_beta.T).sum() / batch.n
        return


    def update_lambda(self, batch):
        """
        Process one minibatch
        """
        # rhot will be between 0 and 1, and says how much to weight
        # the information we got from this mini-batch.

        # E step to update gamma | lambda for mini-batch
        self.do_e_step(batch)

        # Estimate likelihood for current values of lambda.
        scores = self.approx_score(batch)
        if self._verbose > 0:
            print(f"{self._updatect}-th global update. Scores: " + ", ".join(['%.2e'%x for x in scores]))

        # Update global parameters
        rhot = pow(self._tau0 + self._updatect, -self._kappa)
        doc_ratio = float(self._N) / batch.n
        self._lambda = (1-rhot) * self._lambda + \
                       rhot * (doc_ratio * (self._eta + self._sstats) )
        self._Elog_beta = utilt.dirichlet_expectation(self._lambda)
        self._expElog_beta = np.exp(self._Elog_beta)
        self._updatect += 1
        return scores


    def transform(self, X, gamma = None, alpha = None):
        assert X.shape[1] == self._M
        if issparse(X):
            X = X.tocsr()
        n = X.shape[0]
        if gamma is None:
            gamma = np.random.gamma(100., 1./100., (n, self._K))
        if alpha is None:
            alpha = np.broadcast_to(self._alpha, (n, self._K))
        else:
            if len(alpha.shape) == 1:
                alpha = np.broadcast_to(alpha, (n, self._K))
            else:
                assert alpha.shape == (n, self._K), "Invalid alpha input"
        # Run E-step in parallel
        if self._thread <= 1:
            _, gamma = self._update_gamma(X, gamma, alpha)
        else:
            idx_slices = [idx for idx in utilt.gen_even_slices(n, self._thread)]
            with Parallel(n_jobs=self._thread, verbose=max(0, self._verbose - 1)) as parallel:
                result = parallel(delayed(self._update_gamma)(X[idx, :], gamma[idx, :], alpha[idx, :])
                for idx in idx_slices)
            for i, v in enumerate(result):
                gamma[idx_slices[i], :] = v[1]
        gamma = normalize(gamma, norm='l1', axis=1)
        return gamma


    def approx_score(self, batch):
        """
        ELBO
        """
        score_gamma = 0
        score_beta  = 0
        Elog_theta = utilt.dirichlet_expectation(batch.gamma)
        E_theta = normalize(batch.gamma, norm='l1', axis=1)

        # E[log p(x | theta, beta)]
        score_unit = batch.mtx @ self._Elog_beta.T
        score_unit = np.multiply(E_theta, score_unit/batch.n).sum()

        # E[log p(theta | alpha) - log q(theta | gamma)]
        score_gamma += np.sum((batch.alpha - batch.gamma)*Elog_theta)
        score_gamma += np.sum(gammaln(batch.gamma)) - np.sum(gammaln(np.sum(batch.gamma, 1)))
        score_gamma /= batch.n

        # E[log p(beta | eta) - log q (beta | lambda)]
        score_beta += np.sum((self._eta-self._lambda)*self._Elog_beta)
        score_beta += np.sum(gammaln(self._lambda)) - np.sum(gammaln(np.sum(self._lambda, 1)))

        return((score_unit, score_gamma, score_beta, self._N * (score_unit+score_gamma) + score_beta))

    def coherence_pmi(self, pw, pseudo_ct = 200, top_gene_n = 100):
        """
        Topic coherence (pointwise mutual information)
        """
        topic_pmi = []
        top_gene_n = np.min([top_gene_n, self._M])
        pw = pw / pw.sum()
        for k in range(self._K):
            indx = np.argsort(-self._Elog_beta[k, :])[:top_gene_n]
            b = np.exp(self._Elog_beta[k, indx] )
            b = np.clip(b, 1e-6, 1.-1e-6)
            w = 1. - np.power(1.-pw[indx], pseudo_ct)
            w = w.reshape((-1, 1)) @ w.reshape((1, -1))
            p0 = 1.-np.power(1-b, pseudo_ct)
            p0 = p0.reshape((-1, 1)) @ p0.reshape((1, -1))
            pmi = np.log(p0) - np.log(w)
            np.fill_diagonal(pmi, 0)
            pmi = np.round(pmi.mean(), 3)
            topic_pmi.append(pmi)
        return topic_pmi



import sys, os, copy, gzip, logging
import numpy as np
import pandas as pd
from scipy.sparse import *
from dataclasses import dataclass
from sklearn.neighbors import BallTree

from ficture.loaders.data_loader import factor_space_stream

@dataclass
class BlockProfile:
    weight: float
    xst: float
    yst: float
    xed: float
    yed: float
    profile: np.ndarray # Unnormalized

class SlidingPosteriorCount:

    def __init__(self, pixel_reader, index_axis, key, factor_file, ft_dict,\
                 size_um, slide_step, mu_scale,\
                 precision = 1, radius = 3, init_pos = 0, debug = 0) -> None:
        self.pixel_reader = pixel_reader
        self.index_axis = index_axis
        self.key = key
        self.ft_dict = ft_dict
        self.size_um = int(size_um / precision) * precision
        self.slide_step = slide_step
        self.mu_scale = mu_scale
        self.precision = precision
        self.radius = radius
        self.block_size = self.size_um / self.slide_step
        self.left_over = pd.DataFrame()
        self.ymin = 0
        self.ymax = 0
        self.file_is_open = True
        self.block_ct_min = 100
        self.blocks = {}
        self.windows = {}
        self.window_centers = []
        self.ref = None
        self.factor_map = factor_space_stream(file=factor_file, index_axis=index_axis, init_pos = init_pos)
        self.K = self.factor_map.K
        self.M = len(self.ft_dict)
        self.debug = debug

    def update_reference(self, st, ed):
        """
        Update the reference to cover from st to ed
        """
        ymin = st - self.size_um
        ymax = ed + self.size_um
        self.add_block(ymin, ymax)
        self.remove_block(ymin)
        self.block_to_windows()
        self.ref = BallTree(np.array([x[:2] for x in self.window_centers]))

    def query(self, x, y):
        """
        Query the nearest window to (x, y)
        Return the distance and the window profile
        """
        pt = np.array([[x, y]])
        dist, indx = self.ref.query(pt, 1)
        k = self.window_centers[indx[0][0]][2]
        return dist[0][0], self.windows[k].profile

    def block_to_windows(self):
        """
        Sum over blocks in each window
        """
        self.window_centers = []
        for k, v in self.blocks.items():
            weight = 0
            beta = np.zeros((self.K, self.M))
            n_block = 0
            for i in range(self.slide_step):
                for j in range(self.slide_step):
                    u = self.blocks.get((k[0] + i, k[1] + j))
                    if u is None:
                        continue
                    beta += u.profile
                    weight += u.weight
                    n_block += 1
            xst = k[0] * self.block_size
            yst = k[1] * self.block_size
            xed = xst + self.size_um
            yed = yst + self.size_um
            xct = (xst + xed) / 2
            yct = (yst + yed) / 2
            self.window_centers.append([xct, yct, k])
            self.windows[k] = BlockProfile(weight = weight, xst = xst, yst = yst, xed = xed, yed = yed, profile = beta)
            if self.debug > 1:
                print(f"Created window {k} with {n_block} blocks, total weight {weight}")

    def add_block(self, ymin, ymax):
        """
        Read data until beyond ymax
        """
        df = copy.copy(self.left_over)
        ymin = ymin / self.mu_scale
        self.block_indx_max = int(ymax / self.block_size) + 1
        ymax = self.block_indx_max * self.block_size
        while self.ymax < ymax:
            try:
                chunk = next(self.pixel_reader)
                chunk = chunk.loc[(chunk[self.index_axis] > ymin) & (chunk[self.key] > 0), :]
                if chunk.shape[0] == 0:
                    continue
                chunk.X *= self.mu_scale
                chunk.Y *= self.mu_scale
                df = pd.concat([df, chunk])
                self.ymax = df[self.index_axis].max()
                if self.debug:
                    print(f"Current cursor: {self.ymax}")
            except StopIteration:
                self.file_is_open = False
                break
        df = df.loc[df.gene.isin(self.ft_dict), :]
        # only process blocks completely within ymax
        left = df[self.index_axis] >= ymax
        self.left_over = copy.copy(df.loc[left, :])
        if self.debug:
            print(f"Read data entries {df.shape[0]}, leftover {self.left_over.shape[0]}")
        df = df.loc[~left, :]
        if df.shape[0] == 0:
            return
        df.X = (df.X / self.precision).astype(int)
        df.Y = (df.Y / self.precision).astype(int)
        df['j'] = list(zip(df.X, df.Y))
        df = df.groupby(by = ['j', 'gene']).agg({self.key: sum}).reset_index()
        brc = df.groupby(by = ['j']).agg({self.key: sum}).reset_index()
        brc['X'] = brc.j.map(lambda x: x[0]) * self.precision
        brc['Y'] = brc.j.map(lambda x: x[1]) * self.precision
        x = (brc.X / self.block_size).astype(int).values
        y = (brc.Y / self.block_size).astype(int).values
        brc['w_id'] = list(zip(x, y))

        # count total reads per block
        w_ct = brc.groupby(by = ['w_id']).agg({self.key: sum}).reset_index()
        # keep only blocks with more than 100 reads
        w_id = set(w_ct.loc[w_ct[self.key] > self.block_ct_min, 'w_id'].values)
        w_id = {x : np.array([x[0],x[0]+1,x[1],x[1]+1]) * self.block_size for x in w_id}
        if self.debug:
            print(f"Total blocks: {brc.w_id.nunique()}, kept blocks: {len(w_id)}")
        brc = brc.loc[brc.w_id.isin(w_id), :]

        # Make DGE
        df = df.loc[df.j.isin(brc.j.values), :]
        barcode_kept = list(brc.j.values)
        bc_dict = {x:i for i,x in enumerate( barcode_kept ) }
        indx_row = [bc_dict[x] for x in df['j']]
        indx_col = [self.ft_dict[x] for x in df['gene']]
        mtx = coo_array((df[self.key].values, (indx_row, indx_col)), shape=(len(barcode_kept), self.M)).tocsr() # (N, M)
        if self.debug:
            print(f"Made DGE: {mtx.shape}")

        # find factor loading of pixels
        theta = self.factor_map.impute_factor_loading(pos = brc.loc[:, ["X", "Y"]].values, k = 1, radius = self.radius, include_self = True) # (N, K)
        # factor specific DGE by block
        for w, v in w_id.items():
            indx = brc.loc[brc.w_id == w, 'j'].map(bc_dict).values
            weight = brc.loc[brc.w_id == w, self.key].sum()
            beta = theta[indx, :].T @ mtx[indx, :] # (K, M)
            self.blocks[w] = BlockProfile(weight = weight, xst = v[0], yst = v[2], xed = v[1], yed = v[3], profile = beta)
            if self.debug > 1:
                print(f"Add block {w} at {v[0]}, {v[2]}, with weight {weight}")

    def remove_block(self, ymin):
        """
        Remove blocks that are below ymin
        """
        for k, v in self.blocks.items():
            if v.yed < ymin/self.block_size:
                self.blocks.pop(k, None)
